#!/usr/bin/python3

import copy
import json
import os
import re
import sys
import yaml
from importlib.resources import files

import editor

from collections import OrderedDict
from jira.exceptions import JIRAError
from referencing import Registry
import jsonschema

from jirate.args import ComplicatedArgs, GenericArgs
from jirate.jboard import JiraProject, get_jira
from jirate.decor import md_print, pretty_date, hbar_under, hbar, hbar_over, nym, vsep_print, parse_params, truncate, render_matrix, comma_separated
from jirate.decor import issue_link_string, link_string
from jirate.decor import pretty_print  # NOQA
from jirate.decor import EscapedString
from jirate.config import get_config, yaml_dump
from jirate.jira_fields import apply_field_renderers, render_issue_fields, max_field_width, render_field_data, jirate_field
from jirate.template_vars import apply_values
from jirate.rqcache import RequestCache


# Prevent case/typos/etc.
_subtask = 'Sub-task'


def move(args):
    if args.user:
        args.project.assign(args.src, args.user)
    if args.mine:
        args.project.assign(args.src, 'me')
    ret = args.project.move(args.src, args.target)
    if ret:
        print('Moved', [issue.key for issue in ret], 'to', args.target)
        return (0, False)
    print('No issue(s) moved')
    return (1, False)


def close_issues(args):
    ret = 0
    close_args = {}
    if args.resolution:
        close_args['resolution'] = args.resolution

    close_issues = []
    if args.subtasks:
        for issue_key in args.target:
            # Close subtasks first, then main issue
            issue = args.project.issue(issue_key)
            if issue.key in close_issues:
                continue
            issue_subtasks = [subtask['key'] for subtask in issue.raw['fields']['subtasks']]
            close_issues.extend(issue_subtasks)
            close_issues.append(issue.key)
    else:
        close_issues = args.target

    for issue in close_issues:
        if not args.project.close(issue, **close_args):
            ret = 1
    return (ret, False)


def parse_field_widths(field_string, allowed_fields=None, ignore_fields=None, starting_fields=None):
    if not starting_fields:
        fields = OrderedDict()
    else:
        fields = starting_fields

    raw_fields = parse_params(field_string)
    for field in raw_fields:
        if ':' in field:
            val = field.split(':')
            field = val[0]
            maxlen = int(val[1])
        else:
            maxlen = 0
        if ignore_fields and field in ignore_fields:
            continue
        if allowed_fields and field not in allowed_fields:
            continue
        fields[field] = maxlen
    return fields


def _reorder_issues(issue_list):
    ret = []
    issues_by_key = OrderedDict()

    # 1. Scrape all issues by key
    for issue in issue_list:
        issues_by_key[str(issue.key)] = issue

    # 2. Reorder
    #    a. group subtasks under parent task
    #    b. orphan subtasks are left in rank order
    for issue in issue_list:
        if str(issue.fields.issuetype) == _subtask:
            # For subtasks, if their parent is not in the list,
            # append now to preserve rank order
            if str(issue.fields.parent) not in issues_by_key:
                ret.append(issue)
            continue
        ret.append(issue)

        # There's a bug in JIRA API. If you get:
        #    issue/ISSUE-123
        # ... the subtasks in the JSON data are in order according to the
        # JIRA web UI.  If, however, you have ISSUE-123 as part of the
        # result of a search, the subtasks' ordering according to the UI
        # is not preserved
        issue_subtasks = [subtask['key'] for subtask in issue.raw['fields']['subtasks']]
        for subtask in issue_subtasks:
            if subtask in issues_by_key:
                ret.append(issues_by_key[subtask])

    return ret


def print_issues_by_field(issue_list, args=None, exclude_fields=[]):
    # TODO: sort by column
    fields = OrderedDict({'key': 0})
    ignore_fields = ['key']
    ignore_fields.extend(exclude_fields)
    fields = parse_field_widths(args.fields, ignore_fields=ignore_fields, starting_fields=fields)
    if args.format != 'csv':
        subtask_prefix = EscapedString('↳ ')
        subtask_error = EscapedString('‼ ')
    else:
        subtask_prefix = EscapedString('')
        subtask_error = EscapedString('')

    if not args.compact:
        args.compact = args.project.get_user_data('compact_output')

    output = []
    raw_fields = list(fields.keys())
    output.append(list(truncate(key, fields[key]) for key in fields))
    del fields['key']

    found_fields = []
    issue_list = _reorder_issues(issue_list)

    # To show subtasks under parents, we need to know all the issues we
    # have so we can see if the parent task is in the list
    issue_keys = [str(issue.key) for issue in issue_list]

    for issue in issue_list:
        if args and hasattr(args, 'status') and args.status:
            if nym(issue.field('status')['name']) != nym(args.status):
                continue
        row = []
        key_string = issue_link_string(issue.key, args.project.jira.server_url)
        # Above, we reordered a subtask to be under its parent task in subtask
        # order according to the parent issue. Here, we want to show a visual break
        # to show this is associated with the above non-Subtask
        if str(issue.fields.issuetype) == _subtask and str(issue.fields.parent) in issue_keys:
            # No f-magic for EscapedString, so we must concatenate
            if issue.raw['fields']['status']['statusCategory']['name'] != 'Done' and issue.raw['fields']['parent']['fields']['status']['statusCategory']['name'] == 'Done':
                key_string = subtask_error + key_string
            else:
                key_string = subtask_prefix + key_string
        row.append(key_string)
        for field in fields:
            # See if it's a user-defined one first, as optimization
            real_key = jirate_field(field)
            if real_key:
                field_key = field
            else:
                # See if it's an alias for a field in jira
                # (then real_key and field_key are the same)
                field_key = args.project.field_to_id(field)
                if not field_key:
                    row.append('N/A')
                    continue
                real_key = field_key
            try:
                raw_fv = issue.field(real_key)
            except AttributeError:
                row.append('N/A')
                continue
            fk, fv = render_field_data(field_key, issue.raw['fields'], True, args.project.allow_code)
            if fk:
                val = fv
            else:
                val = raw_fv
            if val is None:
                val = ''
            row.append(truncate(val, fields[field]))
            if args.compact and val == '':
                continue
            if field not in found_fields:
                found_fields.append(field)
        output.append(row)

    header = not (args.format in ['csv'])
    if header:  # Fancy output: save real estate by removing columns
        delta = list(set(list(fields.keys())) - set(found_fields))
        for kill in delta:
            try:
                column = raw_fields.index(kill)
                raw_fields.pop(column)
            except ValueError:
                print(f'Bug: Tried to remove nonexistent column {kill}?')
                continue
            for row in output:
                row.pop(column)

    lines = render_matrix(output, fmt=args.format, header=header)
    return lines


def print_issues_by_state(issue_list, args=None):
    states = {}
    printed = 0

    for issue in issue_list:
        cstatus = issue.raw['fields']['status']['name']
        if cstatus not in states:
            states[cstatus] = []
        states[cstatus].append(issue)

    for key in states:
        if args and hasattr(args, 'status') and args.status and nym(key) != nym(args.status):
            continue
        hbar_under(key)
        for issue in states[key]:
            printed = printed + 1
            issue_info = EscapedString('  ') + issue_link_string(issue.key, args.project.jira.server_url)
            print(issue_info, end=' ')
            if args and hasattr(args, 'labels') and args.labels:
                print_labels(issue.raw, prefix='')
            print(issue.raw['fields']['summary'])
        print()
    return printed


def print_keys(issue_list):
    for issue in issue_list:
        print(issue.key)
    return False


def print_issues(issue_list, args=None, exclude_fields=[]):
    footer = (args.format in ('default'))
    total = len(issue_list)

    if not issue_list:
        print('No matching issues')
        footer = False
    elif not args:
        total = print_issues_by_state(issue_list, args)
    elif hasattr(args, 'quiet') and args.quiet:
        print_keys(issue_list)
        footer = False
    elif hasattr(args, 'fields') and args.fields is not None:
        total = print_issues_by_field(issue_list, args, exclude_fields)
    else:
        fields = args.project.get_user_data('default_fields')
        if fields:
            setattr(args, 'fields', fields)
            total = print_issues_by_field(issue_list, args, exclude_fields)
        else:
            total = print_issues_by_state(issue_list, args)
    if footer and total is not None:
        hbar_over(str(total) + ' result(s)')
    return True


def print_users(users):
    matrix = [['Name', 'User Name', 'Email Address']]

    for user in users:
        matrix.append([user.displayName, user.name, user.emailAddress])
    render_matrix(matrix)


# Returns the search name and fields if provided
def find_search(name, search_info):
    if name not in search_info:
        return (None, None)
    item = search_info[name]
    if isinstance(item, str):
        return (item, None)
    elif not isinstance(item, dict) or 'query' not in item:
        return (None, None)

    fields = None
    if 'fields' in item:
        fields = item['fields']
    return (item['query'], fields)


def search_jira(args):
    if args.user:
        users = args.project.search_users(args.user)
        if not users:
            print(f'No users match "{args.user}"')
            return (1, False)
        print_users(users)
        return (0, False)

    named = args.named_search
    fields = None
    if not args.text and not named:
        named = 'default'
    if named:
        searches = args.project.get_user_data('searches')
        (search_query, fields) = find_search(named, searches)
        if not search_query:
            print(f'No search configured: {named}')
            return (1, False)
        if args.text:
            additional = ' '.join(args.text)
            search_query = f'({additional}) AND {search_query}'

        # Field priority:
        # 1. cli fields
        # 2. per-search fields
        # 3. global fields
        if (not hasattr(args, 'fields') or args.fields is None) and fields:
            setattr(args, 'fields', fields)
        ret = args.project.search_issues(search_query)
    else:
        search_query = ' '.join(args.text)
        if args.raw:
            ret = args.project.search_issues(search_query)
        else:
            ret = args.project.search(search_query)

    # JIRA's text search borders on useless.
    # Prune any issues from output where the regex does not
    # match supplied field
    stripped = []
    if args.prune_regex:
        field = args.prune_regex[0]
        fid = args.project.field_to_id(field)
        regex = args.prune_regex[1]
        for issue in ret:
            val = issue.field(field)
            try:
                if val and re.search(regex, val):
                    stripped.append(issue)
                continue
            except TypeError:
                pass
            (_, val) = render_field_data(fid, issue.raw['fields'], True, args.project.allow_code)
            if val and re.search(regex, val):
                stripped.append(issue)
        ret = stripped

    if not ret:
        return (127, False)
    print_issues(ret, args)
    return (0, False)


def list_issues(args):
    # check for verbose
    if args.unassigned:
        userid = None
    elif args.user:
        userid = args.user
    else:
        userid = 'me'

    issues = args.project.list(status=args.status, userid=userid, all_issues=args.all)
    print_issues(issues, args)
    return (0, True)


def list_link_types(args):
    ltypes = args.project.link_types()
    matrix = [['Outward', 'Inward']]
    for lt in ltypes:
        matrix.append([lt.outward, lt.inward])
    render_matrix(matrix)
    return (0, True)


def list_states(args):
    states = args.project.states()
    namelen = 0
    for name in states:
        namelen = max(namelen, len(name))
    for name in states:
        vsep_print(None, 0, name, namelen, states[name]['name'])
    return (0, False)


def list_issue_types(args):
    issue_types = args.project.issue_types
    for itype in issue_types:
        print('  ', itype.name)
    return (0, False)


def issue_fields(args):
    if args.issue:
        issue = args.project.issue(args.issue)
        if not issue:
            print('No such issue:', args.issue)
            return (1, False)
        fields = args.project.fields(issue.raw['key'])
    elif args.type:
        md = args.project.issue_metadata(args.type)
        if md:
            fields = md['fields']
        else:
            print(f'No metadata for {args.type}')
            return (1, False)

    # Remove things we set elsewhere
    for field in ('assignee', 'issuelinks', 'comment'):
        if field in fields:
            del fields[field]

    # Remove things we don't support setting
    for field in ('issuetype', 'attachment', 'reporter', 'project'):
        if field in fields:
            del fields[field]

    display = False
    try:
        if args.operation:
            pass
    except AttributeError:
        display = True

    if display:
        matrix = [['Field', 'Allowed Values']]
        for field in fields:
            if field.startswith('customfield_'):
                fname = nym(fields[field]['name'])
            else:
                fname = nym(field)
            fvalue = ''
            if 'allowedValues' in fields[field]:
                values = []
                for val in fields[field]['allowedValues']:
                    if 'archived' in val and val['archived']:
                        continue
                    if 'name' in val:
                        values.append(val['name'])
                    elif 'value' in val:
                        values.append(val['value'])
                    else:
                        values.append(val['id'])
                fvalue = comma_separated(values)
            matrix.append([fname, fvalue])
        render_matrix(matrix)
        return (0, False)

    # Update a field
    field_name = args.name
    field_id = args.project.field_to_id(field_name)
    if field_id not in fields:
        raise ValueError(f'Update to {field_name} ({field_id}) is not allowed at this point')

    schema = fields[field_id]['schema']
    if schema['type'] == 'user':
        value = args.project.get_user(args.values[0])
    elif schema['type'] == 'array':
        # Special case: Resolve users before passing down
        if schema['items'] == 'user':
            value = [args.project.get_user(user) for user in args.values]
        else:
            value = args.values
    else:
        value = ' '.join(args.values)

    op = args.operation
    # Substitution only works on 'set' capable fields for now
    if args.operation == 'sub':
        op = 'set'
        if len(args.values) < 2:
            raise ValueError('Substitution requires an old value and a new value')
        oldval = args.values[0]
        newval = args.values[1]
        (_, orig_value) = render_field_data(field_id, issue.raw['fields'], True, args.project.allow_code)
        value = orig_value.replace(oldval, newval)

    try:
        issue.update_field(field_name, value, op, fields)
    except (AttributeError, ValueError) as e:
        print(e)
        return (1, False)
    return (0, False)


def split_issue_text(text):
    lines = text.split('\n')
    name = lines[0]
    desc = ''
    if not len(name):
        return (None, None)
    lines.pop(0)
    while len(lines) and lines[0] == '':
        lines.pop(0)
    if len(lines):
        desc = '\n'.join(lines)
    return (name, desc)


def new_issue(args):
    desc = None

    if args.text:
        name = ' '.join(args.text)
    else:
        text = editor()
        name, desc = split_issue_text(text)
        if name is None:
            print('Canceled')
            return (1, False)

    issue = args.project.new(name, desc, issue_type=args.type)
    if args.quiet:
        print(issue.raw['key'])
    else:
        print_issue(args.project, issue, False)
    return (0, True)


def print_creation_fields(metadata):
    nlen = 5
    fields = metadata['fields']
    ignore_fields = ('Reporter', 'Project', 'Issue Type')
    for field in fields:
        if fields[field]['name'] in ignore_fields:
            continue
        if field.startswith('customfield_'):
            nlen = max(nlen, len(nym(fields[field]['name'])))
        else:
            nlen = max(nlen, len(nym(field)))

    matrix = [['Field', 'R', 'Allowed Values']]

    for field in fields:
        if fields[field]['name'] in ignore_fields:
            continue
        if field.startswith('customfield_'):
            fname = nym(fields[field]['name'])
        else:
            fname = nym(field)
        fvalue = ''
        if fields[field]['required']:
            req = '*'
        else:
            req = ' '
        if 'allowedValues' in fields[field]:
            values = []
            for val in fields[field]['allowedValues']:
                if 'archived' in val and val['archived']:
                    continue
                if 'name' in val:
                    values.append(val['name'])
                elif 'value' in val:
                    values.append(val['value'])
                else:
                    values.append(val['id'])
            fvalue = comma_separated(values)
        matrix.append([fname, req, fvalue])
    render_matrix(matrix)


def _metadata_by_type(project, issuetype):
    try:
        metadata = project.issue_metadata(issuetype)
    except JIRAError as e:
        if 'text: Issue Does Not Exist' in str(e):
            print('The createmeta API does not exist on this JIRA instance.')
        else:
            print(e)
        return (1, False)

    if not metadata:
        list_issue_types(project.issue_types)
        raise ValueError(f'Invalid issue type: {issuetype}')
    return metadata


# new_issue is way too easy. Let's make it *incredibly* complicated!
def create_issue(args):
    desc = None
    issuetype = args.type if args.type else 'Task'

    # Do sanity check before hitting the JIRA API
    if len(args.args) % 2 == 1:
        print('Incorrect number of arguments (not divisible by 2)')
        return (1, False)

    metadata = _metadata_by_type(args.project, issuetype)

    values = {}
    argv = copy.copy(args.args)
    if len(argv) == 0:
        print_creation_fields(metadata)
        return (0, False)

    while len(argv):
        key = argv.pop(0)
        value = argv.pop(0)
        values[key] = value

    # Bug - error checking isn't done until later, but we need a
    # summary or the below code blows up.  So, you might pop up
    # $EDITOR only to find out later that you used an invalid
    # issue type.
    if 'summary' not in values:
        text = editor()
        name, desc = split_issue_text(text)
        if name is None:
            print('Canceled')
            return (1, False)
        values['summary'] = name
        if desc:
            values['description'] = desc

    values['issuetype'] = metadata['name']
    values['project'] = args.project.project_name

    issue = args.project.create(metadata['fields'], **values)
    if args.quiet:
        print(issue.raw['key'])
    else:
        print_issue(args.project, issue, False)
    return (0, True)


def _parse_creation_args(issue_data, required_fields=None, reserved_fields=None, translate_fields=None, start_vals=None):
    ret = {}
    if start_vals:
        ret = start_vals

    for key in issue_data:
        actual_key = key
        if translate_fields and key in translate_fields:
            actual_key = translate_fields[key]
        if not reserved_fields or actual_key not in reserved_fields:
            ret[actual_key] = issue_data[key]
    missing = []
    if required_fields:
        for key in required_fields:
            if key not in ret:
                missing.append(key)
        if missing:
            raise ValueError(f'Missing required fields: {missing}')
    return ret


def create_from_template(args):
    with open(args.template_file, 'r') as yaml_file:
        template = yaml_file.read()

    values = {}
    # Always render. there should be defaults.
    if args.vars:
        if len(args.vars) % 2 != 0:
            raise ValueError('Variable/value list is not divisible by 2')

        # Someone set up us the render
        argv = copy.copy(args.vars)
        while len(argv):
            key = argv.pop(0)
            value = argv.pop(0)
            values[key] = value

    interactive = sys.stdin.isatty() and not args.non_interactive

    # template_output is the raw text with jinja2 variable substitution
    # completed, not a yaml structure
    template_output = apply_values(template, values, interactive)

    template = yaml.safe_load(template_output)

    try:
        _validate_template(args.project, template)
    except jsonschema.exceptions.ValidationError as e:
        print(f"Provided template file is not valid: {args.template_file}")
        raise e

    if args.dry_run:
        print(template_output)
        return (0, True)

    all_filed = _create_from_template(args, template)

    # Need to refresh to that issues get re-fetched to include subtasks
    # TODO: Have subtask() update parent issues in _config['issue_map']
    args.project.delete_issue_map()
    for filed in all_filed:
        if args.quiet:
            if 'subtasks' in filed:
                print(filed['parent'] + ': ' + ', '.join(filed['subtasks']))
            else:
                print(filed['parent'])
        else:
            print_issue(args.project, args.project.issue(filed['parent']), False)
    return (0, True)


def _create_from_template(args, template):
    # Cache for issue createmeta information
    metadata_by_type = {}

    # TODO: consider using Jira's bulk issue creation
    # TODO: support reading arbitrary fields from the template
    all_filed = []  # We keep issue keys here because we'll need to refresh anyway

    existing_issue = None
    if args.apply:
        # Sanity checks:
        # 1. Issue exists
        existing_issue = args.project.issue(args.apply)
        if not existing_issue:
            raise ValueError(f'Cannot apply template to nonexisting issue {args.apply}')
        # 2. Template's issues attribute length is 1
        if len(template['issues']) != 1:
            raise ValueError('Undefined request: template is for multiple issues')

    projects = {}
    projects[args.project.project_name] = args.project

    for raw_issue in template['issues']:
        issue = {args.project.field_to_id(name): value for name, value in raw_issue.items()}
        reserved_fields = ['subtasks']
        required_fields = ['summary']
        if 'project' in issue:
            pname = issue['project']
        else:
            pname = args.project.project_name
        if pname not in metadata_by_type:
            metadata_by_type[pname] = {}
        if pname not in projects:
            projects[pname] = JiraProject(args.project.jira, pname, readonly=False, allow_code=args.project.allow_code)

        creation_fields = _parse_creation_args(issue, required_fields, reserved_fields)

        # Cache all metadata now (so we can debug subtask creation if needed.
        issuetype = creation_fields['issuetype']
        if issuetype not in metadata_by_type:
            metadata_by_type[pname][issuetype] = _metadata_by_type(projects[pname], issuetype)
        if 'subtasks' in issue and issue['subtasks']:
            metadata_by_type[pname][_subtask] = _metadata_by_type(projects[pname], _subtask)
        metadata = metadata_by_type[pname][issuetype]

        filed = {}
        if existing_issue:
            # Do not mess with issuetype if we're applying to existing issue
            del creation_fields['issuetype']
            existing_issue.update(**creation_fields)
            parent = existing_issue
        else:
            parent = projects[pname].create(metadata['fields'], **creation_fields)
        filed['parent'] = parent.key

        # Apply subtasks - but only to a parent which does not already have any
        # subtasks
        if ('subtasks' in issue and issue['subtasks']) and ('subtasks' not in parent.raw['fields'] or not parent.raw['fields']['subtasks']):
            # Set once
            filed['subtasks'] = []
            for subtask in issue['subtasks']:
                metadata = metadata_by_type[pname][_subtask]
                reserved_fields = ['subtasks', 'issuetype', 'parent']
                # required_fields are the same
                start_fields = {'issuetype': _subtask, 'project': pname}
                start_fields['parent'] = parent.key
                creation_fields = _parse_creation_args(subtask, required_fields, reserved_fields, start_vals=start_fields)

                child = projects[pname].create(metadata['fields'], **creation_fields)
                filed['subtasks'].append(child.key)
        all_filed.append(filed)

    return all_filed


def _validate_template(project, template):
    for i, issue in enumerate(template['issues']):
        template['issues'][i] = {project.field_to_id(name): value for name, value in issue.items()}

    schema_dir = files('jirate').joinpath('schemas')
    schemas = {}
    for schema_file in ('template.yaml', 'issue.yaml', 'subtask.yaml'):
        schemas[f"jirate:{schema_file}"] = yaml.safe_load(
            schema_dir.joinpath(schema_file).read_text())

    registry = Registry().with_contents([(k, v) for k, v in schemas.items()])
    validator = jsonschema.Draft202012Validator(schema=schemas['jirate:template.yaml'], registry=registry)

    # Will raise a ValidationError with details on what failed:
    validator.validate(template)
    return True


def validate_template(args):
    with open(args.template_file, 'r') as yaml_file:
        template = yaml.safe_load(yaml_file)

    _validate_template(args.project, template)

    # If we get here it means validation succeeded.
    print(f"Template {args.template_file} is valid.")
    return (0, True)


def generate_template(args):
    template = {}
    issue = args.project.issue(args.issue_id)
    if not issue:
        print('No such issue:', args.issue_id)
        return (1, False)

    # TODO: allow customizing allow_fields in the config
    template = _generate_template(issue.raw['fields'], args.project.field_to_alias, args.project.issue, args.all_fields)

    print(yaml_dump({'issues': [template]}))
    return (0, True)


def _generate_template(raw_issue, translate_fields, fetch_issue=None, all_fields=False, allow_fields=None):
    template = _serialize_issue(raw_issue, translate_fields, fetch_issue)

    if not all_fields:
        if allow_fields is not None:
            # Ensure these are in alias form, if set. Makes writing configs easier.
            allow_fields = [translate_fields(field) for field in allow_fields]
        template = _trim_template(template, allow_fields)
    return template


def _serialize_issue(raw_issue, translate_fields, fetch_issue=None):
    """
    Serializes raw JIRA issue fields for use in Jirate templates.

    Parameters:
      raw_issue: issue fields (typically issue.raw['fields']) (dict)
      translate_fields: field name translator (typically a JiraProject's field_to_alias method) (function)
      fetch_issue: issue data fetcher, used to get subtask data (typically a JiraProject's issue method) (function)

    Returns:
      Processed issue dict ready to be dumped as part of a YAML template (dict)
    """
    template = {}
    _subtasks = 'subtasks'  # ID form

    for field, value in raw_issue.items():
        if field == _subtasks and value:
            # The parent issue only includes some of the subtask data, so we need to fetch the rest.
            template[translate_fields(_subtasks)] = []
            for subtask_stub in raw_issue[_subtasks]:
                subtask = fetch_issue(subtask_stub['key'])
                template[translate_fields(_subtasks)].append(_serialize_issue(subtask.raw['fields'], translate_fields))
        else:
            _, template[translate_fields(field)] = render_field_data(field, raw_issue, as_object=True)

    return template


def _trim_template(template, allow_fields=None):
    """
    Returns a trimmed template containing only the fields from the input template that
    are both in an allowlist and initialized (not empty).
    """
    # Field names are already translated to aliases so we don't need to
    # deal with customfield_ nonsense here.
    if allow_fields is None:
        allow_fields = [
            'fixversions',
            'priority',
            'issue_type',
            'summary',
            'description',
            'story_points',
            'components',
            'versions',
            'labels',
            'sub_tasks',
        ]
    _subtasks = 'sub_tasks'  # Alias form, see above comment

    trimmed_fields = {}
    for field, value in template.items():
        if field not in allow_fields:
            continue
        # We can't just check truthiness because False or 0 could be a legitimate value
        if value is None:
            continue
        if (isinstance(value, list) or isinstance(value, dict)) and len(value) == 0:
            continue

        # We want the field and it isn't empty-ish
        if field == _subtasks and len(value) > 0:
            trimmed_fields[_subtasks] = [_trim_template(subtask, allow_fields) for subtask in value]
        else:
            trimmed_fields[field] = value
    return trimmed_fields


def new_subtask(args):
    desc = None
    parent_issue = args.project.issue(args.issue_id)
    if not parent_issue:
        print('No such issue:', args.issue_id)
        return (1, False)

    if args.text:
        name = ' '.join(args.text)
    else:
        text = editor()
        name, desc = split_issue_text(text)
        if name is None:
            print('Canceled')
            return (1, False)

    issue = args.project.subtask(parent_issue.raw['key'], name, desc)
    if args.quiet:
        print(issue.raw['key'])
    else:
        print_issue(args.project, issue, False)
    return (0, True)


def link_issues(args):
    left_issue = args.issue_left
    if not left_issue:
        print('No such issue:', args.issue_left)
        return (1, False)
    right_issue = args.issue_right
    if not right_issue:
        print('No such issue:', args.issue_right)
        return (1, False)
    link_name = ' '.join(args.text)

    args.project.link(left_issue, right_issue, link_name)
    return (0, True)


def unlink_issues(args):
    args.project.unlink(args.issue_left, args.issue_right)
    return (0, True)


def link_url(args):
    issue = args.issue
    if not issue:
        print('No such issue:', args.issue)
        return (1, False)
    url = args.url
    text = ' '.join(args.text)
    args.project.attach(issue, url, text)
    return (0, True)


def quote_reply(args):
    if args.text:
        raise ValueError('Can\'t both quote-reply and one-shot comment')

    issue_id = args.issue
    issue = args.project.issue(issue_id)

    if args.reply is True:
        # Find last comment if we didn't pick one to reply to; jira
        # keeps them chronologically according to add date
        last_cmt = issue.raw['fields']['comment']['comments'][-1]
        comment_id = last_cmt['id']
    else:
        comment_id = args.reply

    # Don't query jira server again when we have the data
    comment = None
    for cmt in issue.raw['fields']['comment']['comments']:
        if cmt['id'] == comment_id:
            comment = cmt
            break
    if not comment:
        return (1, False)

    starting_text = f"Quoth [~{comment['author']['name']}] - {pretty_date(comment['updated'])}:\n"
    starting_text = starting_text + '\n'.join(['┃ ' + item for item in comment['body'].strip().split('\n')])
    new_text = editor(starting_text)
    if 'visibility' in comment:
        group_name = comment['visibility']['value']
    else:
        group_name = None
    if not new_text or new_text == starting_text:
        print('Canceled')
        return (0, False)
    args.project.comment(issue_id, new_text, group_name)
    return (0, False)


def comment(args):
    if args.reply:
        return quote_reply(args)

    issue_id = args.issue

    if args.remove:
        comment_id = args.remove
        comment = args.project.get_comment(issue_id, comment_id)
        comment.delete()
        return (0, False)

    if args.edit:
        comment_id = args.edit
        comment = args.project.get_comment(issue_id, comment_id)
        if args.text:
            new_text = ' '.join(args.text)
        else:
            new_text = editor(comment.body)
            if not new_text:
                print('Canceled')
                return (0, False)

        update_args = {'body': new_text}
        update_anyway = False
        if args.group:
            if args.group.lower() == 'all':
                # This took some doing; it's not well-documented.
                # This clears the 'visibility' property of the comment,
                # making it viewable by all users
                update_args['visibility'] = {'identifier': None}
            else:
                update_args['visibility'] = {'type': 'group', 'value': args.group}
            update_anyway = True
        if update_anyway or comment.body != new_text:
            comment.update(**update_args)
        else:
            print('No changes')
        return (0, False)

    if args.group:
        group_name = args.group
    else:
        group_name = None

    if args.text:
        text = ' '.join(args.text)
    else:
        text = editor()

    if not len(text):
        print('Canceled')
        return (0, False)

    args.project.comment(issue_id, text, group_name)
    return (0, False)


def display_comment(server_url, comment, verbose, no_format):
    # Let's get fancy
    commentator_url = f"{server_url}/secure/ViewProfile.jspa?name={comment['updateAuthor']['key']}"
    commentator = link_string(comment['updateAuthor']['displayName'], commentator_url)
    print(commentator, '-', pretty_date(comment['updated']), '• ID:', comment['id'])
    if 'visibility' in comment:
        print('🔒', comment['visibility']['type'], '•', comment['visibility']['value'])
    hbar(20)
    md_print(comment['body'], no_format)
    print()


def print_labels(issue, prefix='Labels: '):
    if 'labels' in issue and len(issue['labels']):
        print(prefix, end='')
        for label in issue['labels']:
            print(label, end=' ')
        print()


def print_issue_links(issue, baseurl=None):
    hbar_under('Issue Links')
    matrix = []
    for link in issue['issuelinks']:
        if 'outwardIssue' in link:
            text = EscapedString(link['type']['outward'] + ' ') + issue_link_string(link['outwardIssue']['key'], baseurl)
            status = link['outwardIssue']['fields']['status']['name']
            desc = link['outwardIssue']['fields']['summary']
        elif 'inwardIssue' in link:
            text = EscapedString(link['type']['inward'] + ' ') + issue_link_string(link['inwardIssue']['key'], baseurl)
            status = link['inwardIssue']['fields']['status']['name']
            desc = link['inwardIssue']['fields']['summary']
        # color_string throws off length calculations
        matrix.append([text, status, desc])
    render_matrix(matrix, False, False)
    print()


def print_remote_links(links):
    hbar_under('External Links')
    matrix = []
    for link in links:
        # color_string throws off length calculations
        lid = str(link.raw['id'])
        text = link.raw['object']['title']
        url = link.raw['object']['url']
        if ret := link_string(text, url):
            matrix.append([lid, ret])
        else:
            matrix.append([lid, text, url])
    render_matrix(matrix, False, False)
    print()


def print_attachments(attachments):
    hbar_under('Attachments')
    matrix = []
    for attachment in attachments:
        # color_string throws off length calculations
        aid = str(attachment['id'])
        text = attachment['filename']
        url = attachment['content']
        if ret := link_string(text, url):
            matrix.append([aid, ret])
        else:
            matrix.append([aid, text, url])
    render_matrix(matrix, False, False)
    print()


# Dict from search or subtask list
def _print_issue_list(header, issues, baseurl=None):
    if not issues:
        return
    hbar_under(header)
    matrix = []
    for task in issues:
        if isinstance(task, str):
            task = issues[task]
        try:
            task_key = issue_link_string(task.key, baseurl)
            status = task.raw['fields']['status']['name']
            summary = task.raw['fields']['summary']
        except AttributeError:
            task_key = issue_link_string(task['key'], baseurl)
            status = task['fields']['status']['name']
            summary = task['fields']['summary']
        matrix.append([task_key, status, summary])
    render_matrix(matrix, False, False)
    print()


def print_subtasks(issue, baseurl=None):
    _print_issue_list('Sub-tasks', issue['subtasks'], baseurl)


def print_eausm_votes(project, issue):
    votes = project.eausm_issue_votes(issue)
    if votes and 'votes' in votes and len(votes['votes']):
        mx = [['Vote', 'Points']]
        for vote in votes['votes']:
            user = project.jira.user_by_key(vote['userId'])
            mx.append([user.displayName, vote['vote']])
        render_matrix(mx)
        print()


def print_issue(project, issue_obj, verbose=False, no_comments=False, no_format=False, allowed_fields=None):
    if verbose:
        # Get votes and watchers in verbose mode
        project.votes(issue_obj)
        project.watchers(issue_obj)

    issue = issue_obj.raw['fields']

    if allowed_fields:
        allowed_fields = parse_field_widths(allowed_fields)
        disp = {}
        allowed_ids = [project.field_to_id(field) for field in allowed_fields]
        for field_id in allowed_ids:
            if field_id in issue:
                disp[field_id] = issue[field_id]
        issue = disp

    key_link = issue_link_string(issue_obj.key, project.jira.server_url)
    lsize = max(len(key_link), max_field_width(issue, verbose, project.allow_code))
    lsize = max(lsize, len('Next States'))

    if 'summary' in issue and issue['summary']:
        vsep_print(' ', 0, key_link, lsize, issue['summary'])
    render_issue_fields(issue, verbose, project.allow_code, lsize)

    if verbose:
        vsep_print(' ', 0, 'ID', lsize, issue_obj.raw['id'])
        vsep_print(None, 0, 'URL', lsize, issue_obj.permalink())
        trans = project.transitions(issue_obj.raw['key'])
        if trans:
            vsep_print(' ', 0, 'Next States', lsize, [tr['name'] for tr in trans])
        else:
            vsep_print(None, 0, 'Next States', lsize, 'No valid transitions; cannot alter status')

    print()
    if 'description' in issue and issue['description']:
        md_print(issue['description'], no_format)
        print()

    if 'issuelinks' in issue and len(issue['issuelinks']):
        print_issue_links(issue, project.jira.server_url)

    # Don't print external links or votes unless in verbose mode since it's
    # another API call?
    if verbose:
        print_eausm_votes(project, issue_obj)

        links = project.remote_links(issue_obj)
        if links:
            print_remote_links(links)

        if 'attachment' in issue and len(issue['attachment']):
            print_attachments(issue['attachment'])

    if 'subtasks' in issue and len(issue['subtasks']):
        print_subtasks(issue, project.jira.server_url)

    if 'issuetype' in issue:
        for megalith in ('Epic', 'Feature'):
            if issue['issuetype']['name'] == megalith:
                ret = project.search_issues(f'"{megalith} Link" = "' + issue_obj.raw['key'] + '"')
                _print_issue_list(f'Issues in {megalith}', ret, project.jira.server_url)

    if no_comments or 'comment' not in issue:
        return
    if issue['comment']['comments']:
        hbar_under('Comments')

        for cmt in issue['comment']['comments']:
            display_comment(project.jira.server_url, cmt, verbose, no_format)


def cat(args):
    issues = []
    for issue_idx in args.issue_id:
        issue = args.project.issue(issue_idx, True)
        if not issue:
            print('No such issue:', issue_idx)
            return (127, False)
        issues.append(issue)

    if not args.fields:
        fields = args.project.get_user_data('issue_fields')
        if fields:
            setattr(args, 'fields', fields)

    if args.format in ['csv']:
        print_issues(issues, args)
        return (0, False)

    if args.no_format:
        no_format = args.no_format
    else:
        no_format = args.project.get_user_data('no_format')

    for issue in issues:
        print_issue(args.project, issue, args.verbose, args.no_comments, no_format, args.fields)
    return (0, False)


def join_issue_text(name, desc):
    if desc:
        return name + '\n\n' + desc
    return name + '\n\n'


def edit_issue(args):
    issue_idx = args.issue

    issue_obj = args.project.issue(issue_idx)
    issue = issue_obj.raw['fields']
    issue_text = join_issue_text(issue['summary'], issue['description'])
    if args.text:
        new_text = ' '.join(args.text)
    else:
        new_text = editor(issue_text)
    if not new_text:
        print('Canceled')
        return (0, False)
    name, desc = split_issue_text(new_text)
    update_args = {}
    if issue['summary'] != name and issue['summary']:
        update_args['summary'] = name
    if issue['description'] != desc:
        update_args['description'] = desc
    if update_args != {}:
        args.project.update_issue(issue_idx, **update_args)
    else:
        print('No changes')
    return (0, False)


def view_issue(args):
    issue_id = args.issue_id
    issue = args.project.issue(issue_id)
    if not issue:
        return (127, False)
    os.system('xdg-open ' + issue.permalink())
    return (0, False)


def assign_issue(args):
    args.project.assign(args.issue_id, args.user)
    return (0, False)


def unassign_issue(args):
    args.project.assign(args.issue_id, 'none')
    return (0, False)


def user_info(args):
    user_info = [GenericArgs(args.project.user)]
    print_users(user_info)
    return (0, False)


def component_ops(args):
    if args.remove:
        # Explicitly one at a time
        ret = args.project.remove_component(args.remove[0])
        return (0, False)

    if args.add:
        if len(args.add) == 1:
            ret = args.project.add_component(args.add[0])
        else:
            ret = args.project.add_component(args.add[0], ' '.join(args.add[1:]))
        if ret:
            return (0, False)
        return (1, False)

    return (1, False)


def call_api(args):
    data = args.project.api_call(args.resource, args.raw)
    if data:
        if args.raw:
            print(data)
        else:
            pretty_print(data)
        return (0, False)
    return (1, False)


def component_list(args):
    comps_data = args.project.components()
    comp_info = {}
    for comp in comps_data:
        if not args.search or re.search(args.search, comp.raw['name']) or not args.quiet and 'description' in comp.raw and re.search(args.search, comp.raw['description']):
            name = comp.raw['name']
            if 'description' not in comp.raw:
                comp_info[name] = {'name': name, 'description': ''}
            else:
                comp_info[name] = {'name': name, 'description': comp.raw['description'].strip()}
    comp_names = sorted(list(comp_info.keys()))

    if args.fields:
        field_arg = args.fields
    else:
        field_arg = args.project.get_user_data('component_fields')

    if field_arg:
        fields = parse_field_widths(field_arg, allowed_fields=['name', 'description'])
    else:
        fields = OrderedDict({'name': 0, 'description': 0})

    if args.quiet:
        for name in comp_names:
            print(name)
    else:
        keys = [fk for fk in fields.keys()]
        matrix = [keys]
        for name in comp_names:
            matrix.append([truncate(comp_info[name][fk], fields[fk]) for fk in keys])
        render_matrix(matrix, fmt=args.format)

    return (0, False)


def sprint_info(args):
    if args.sprint_id:
        search = f'sprint = {args.sprint_id}'
        if not args.all_types:
            search = search + f' and issuetype != {_subtask}'
        if not args.closed:
            search = search + ' and statusCategory != Done'
        if args.new:
            search = search + ' and statusCategory = New'
        if args.raw:
            if args.raw.lower().strip().starswith('order'):
                search = search + f' {args.raw}'
            else:
                search = search + f' and {args.raw}'
        # FIXME: This doesn't allow for field substring ("order" in the
        # "summary" field, for example)
        if not args.raw or 'order' not in args.raw.lower():
            search = search + ' order by rank desc'
        issues = args.project.search_issues(search)
        print_issues(issues, args, exclude_fields=['sprint'])
        return (0, False)

    # General Sprit information
    if args.closed:
        info = args.project.sprint_info(states=['active', 'future', 'closed'])
    else:
        info = args.project.sprint_info()

    board_by_id = {}
    for board in info['boards']:
        # This is a dict, we want the board, not key
        board = info['boards'][board]
        board_by_id[board.id] = board

    matrix = [['name', 'id', 'status', 'start', 'end', 'board']]
    for sprint in info['sprints']:
        sprint = info['sprints'][sprint]
        if sprint.state not in ('active', 'future') and not args.closed:
            continue
        try:
            board_name = board_by_id[sprint.originBoardId]
        except KeyError:
            board_name = '???'
        matrix.append([sprint.name, sprint.id, sprint.state, pretty_date(sprint.startDate), pretty_date(sprint.endDate), board_name])
    if len(info) > 1:
        header = not (args.format in ['csv'])
        render_matrix(matrix, fmt=args.format, header=header)

    return (0, False)


def eausm_vote(args):
    issues = args.project.issues(args.issue_id)
    for issue in issues:
        args.project.eausm_vote_issue(issue, args.vote)
    return (0, False)


def vote(args):
    issues = args.project.issues(args.issue_id)
    for issue in issues:
        if args.remove:
            args.project.jira.remove_vote(issue.key)
        else:
            args.project.jira.add_vote(issue.key)
    return (0, False)


def get_jira_project(project=None, config=None, config_file=None, **kwargs):
    # project: Project key
    # config: dict / pre-read JSON data
    if not config:
        config = get_config(config_file)
    allow_code = False

    if 'jira' not in config:
        print('No JIRA configuration available')
        return None
    jconfig = config['jira']

    if not project:
        if 'default_project' in jconfig and not project:
            project = jconfig['default_project']
        else:
            print('No JIRA project specified')
            return None

    # Allows users to represent custom fields in output.
    # Not recommended to enable.
    if 'here_there_be_dragons' in jconfig:
        if jconfig['here_there_be_dragons'] is True:
            allow_code = True

    # Configure fancy output rendering if specified
    if 'fancy_output' in jconfig and jconfig['fancy_output']:
        import jirate.decor  # NOQA
        jirate.decor.fancy_output = True
        if 'color_shift' in jconfig:
            jirate.decor.color_shift = int(jconfig['color_shift'])

    if not project:
        # Not sure why I used an array here
        project = jconfig['default_project']
    if 'proxies' not in jconfig:
        jconfig['proxies'] = {"http": "", "https": ""}

    if 'cache_expire' in jconfig:
        expire = jconfig['cache_expire']
    else:
        expire = None

    if 'cache_file' in jconfig:
        cache_file = jconfig['cache_file']
    else:
        cache_file = '~/.jirate.cache'

    jira = get_jira(jconfig)
    cache = RequestCache(jira._session, filename=cache_file, expire=expire)
    proj = JiraProject(jira, project, readonly=False, allow_code=allow_code)
    proj.request_cache = cache
    for key in jconfig:
        if key not in ['custom_fields', 'proxies', 'here_there_be_dragons', 'url', 'token', 'default_project', 'proxies']:
            proj.set_user_data(key, jconfig[key])

    if 'custom_fields' in jconfig:
        if isinstance(jconfig['custom_fields'], str):
            proj.custom_fields = get_config(jconfig['custom_fields'])
        else:
            proj.custom_fields = copy.deepcopy(jconfig['custom_fields'])
    else:
        proj.custom_fields = []

    # CLI field parsing definition
    field_info = None
    if 'field' in kwargs and kwargs['field']:
        field_name = kwargs['field'][0]
        field_info = json.loads(kwargs['field'][1])
        field_id = proj.field_to_id(field_name)
        found = False
        for cf in proj.custom_fields:
            if cf['id'] == field_id:
                found = True
                for key in field_info:
                    # TODO update custom field call in jboard?
                    cf[key] = field_info[key]
                break
        if not found:
            # No config present: Create a field definition on the fly
            field_info['id'] = field_id
            field_info['name'] = proj.field_to_human(field_id)
            proj.custom_fields.append(field_info)

    apply_field_renderers(proj.jira.fields(), False)
    if proj.custom_fields:
        reorder = True
        if 'custom_reorder' in jconfig:
            reorder = jconfig['custom_reorder']
            if reorder is not False:
                reorder = True
        apply_field_renderers(proj.custom_fields, reorder)
    return proj


def add_list_options(cmd,
                     fields_help='Display these fields in a table',
                     quiet_help='Only print issue IDs'):
    cmd.add_argument('-f', '--fields', help=fields_help)
    cmd.add_argument('-q', '--quiet', default=False, help=quiet_help, action='store_true')
    cmd.add_argument('--compact', default=False, help='Delete columns with no value set in matrix output', action='store_true')


def create_parser():
    parser = ComplicatedArgs()

    parser.add_argument('-c', '--config', help='Use this config file (instead of ~/.jirate.json)', default=None)
    parser.add_argument('-p', '--project', help='Use this JIRA project instead of default', default=None, type=str.upper)
    parser.add_argument('-f', '--format', help='Use this format for issue list output', default='default', choices=['default', 'csv'], type=str.lower)
    parser.add_argument('--x-format-field', nargs=2, help='Experimental: apply field formatting from the CLI (field, json)', default=None)
    parser.add_argument('--debug', help='Enable debugging', default=False, action='store_true')

    cmd = parser.command('whoami', help='Display current user information', handler=user_info)

    cmd = parser.command('ls', help='List issue(s)', handler=list_issues)
    cmd.add_argument('-U', '--unassigned', action='store_true', help='Display only issues with no assignee.')
    cmd.add_argument('-u', '--user', help='Display only issues assigned to the specific user.')
    cmd.add_argument('-l', '--labels', action='store_true', help='Display issue labels.')
    cmd.add_argument('-a', '--all', action='store_true', help='Display all issues; do not restrict to one project.')
    add_list_options(cmd)

    cmd.add_argument('status', nargs='?', default=None, help='Restrict to issues in this state')

    cmd = parser.command('search', help='Search issue(s)/user(s) with matching text', handler=search_jira)
    cmd.add_argument('-u', '--user', help='Search for user(s) (max)')
    cmd.add_argument('-n', '--named-search', help='Perform preconfigured named search for issues')
    cmd.add_argument('-r', '--raw', action='store_true', help='Perform raw JQL query')
    cmd.add_argument('--prune-regex', nargs=2, help='Prune results by checking named field against regular expression, removing any that do not match')
    add_list_options(cmd)
    cmd.add_argument('text', nargs='*', help='Search text')

    cmd = parser.command('cat', help='Print issue(s)', handler=cat)
    cmd.add_argument('-v', '--verbose', action='store_true', help='Verbose output')
    cmd.add_argument('-N', '--no-comments', action='store_true', default=False, help='Skip comments')
    cmd.add_argument('-n', '--no-format', action='store_true', help='Do not format output using Markdown')
    add_list_options(cmd)
    cmd.add_argument('issue_id', nargs='+', help='Target issue(s)', type=str.upper)

    cmd = parser.command('view', help='Display issue in browser', handler=view_issue)
    cmd.add_argument('issue_id', help='Target issue', type=str.upper)

    parser.command('ll', help='List states available to project', handler=list_states)
    parser.command('lt', help='List issue types available to project', handler=list_issue_types)
    parser.command('link-types', help='Display link types', handler=list_link_types)

    cmd = parser.command('assign', help='Assign issue', handler=assign_issue)
    cmd.add_argument('issue_id', help='Target issue', type=str.upper)
    cmd.add_argument('user', help='Target assignee')
    # cmd.add_argument('users', help='First is assignee; rest are watchers (if none, assign to self)', nargs='*')

    cmd = parser.command('unassign', help='Remove assignee from issue', handler=unassign_issue)
    cmd.add_argument('issue_id', help='Target issue', type=str.upper)

    cmd = parser.command('mv', help='Move issue(s) to new state', handler=move)
    cmd.add_argument('-m', '--mine', action='store_true', help='Also assign to myself')
    cmd.add_argument('-u', '--user', help='Also assign to user')
    cmd.add_argument('src', metavar='issue', nargs='+', help='Issue key(s)')
    cmd.add_argument('target', help='Target state')

    cmd = parser.command('new', help='Create a new issue', handler=new_issue)
    cmd.add_argument('-t', '--type', default='task', help='Issue type (project-dependent)')
    cmd.add_argument('-q', '--quiet', default=False, help='Only print new issue ID after creation (for scripting)', action='store_true')
    cmd.add_argument('text', nargs='*', help='Issue summary')

    cmd = parser.command('create', help='Create a new issue (advanced)', handler=create_issue)
    cmd.add_argument('-t', '--type', default='task', help='Issue type (project-dependent; default is "Task")')
    cmd.add_argument('-q', '--quiet', default=False, help='Only print new issue ID after creation (for scripting)', action='store_true')
    cmd.add_argument('args', nargs='*', help='field1 "value1" field2 "value2" ... fieldN "valueN".  If none specified, print out possible fields and (if applicable) allowed values. Fields marked with an asterisk are required.')

    cmd = parser.command('subtask', help='Create a new subtask', handler=new_subtask)
    cmd.add_argument('-q', '--quiet', default=False, help='Only print subtask ID after creation (for scripting)', action='store_true')
    cmd.add_argument('issue_id', help='Parent issue', type=str.upper)
    cmd.add_argument('text', nargs='*', help='Subtask summary')

    cmd = parser.command('link', help='Create link between two issues', handler=link_issues)
    cmd.add_argument('issue_left', help='First issue', type=str.upper)
    cmd.add_argument('text', nargs='+', help='Link text')
    cmd.add_argument('issue_right', help='Second issue', type=str.upper)

    cmd = parser.command('attach', help='Attach a web link to an issue', handler=link_url)
    cmd.add_argument('issue', help='Issue', type=str.upper)
    cmd.add_argument('url', help='URL to attach')
    cmd.add_argument('text', nargs='+', help='URL Description')

    cmd = parser.command('unlink', help='Remove link(s) between issues or an external link', handler=unlink_issues)
    cmd.add_argument('issue_left', help='First issue', type=str.upper)
    cmd.add_argument('issue_right', help='Second issue (or external link ID)', type=str.upper)

    cmd = parser.command('comment', help='Comment (or remove) on an issue', handler=comment)
    cmd.add_argument('-e', '--edit', help='Comment ID to edit')
    cmd.add_argument('-r', '--remove', help='Comment ID to remove')
    cmd.add_argument('-q', '--reply', nargs='?', help='Comment ID to quote and reply', default=False, const=True)
    cmd.add_argument('-g', '--group', help='Specify comment group visibility')
    cmd.add_argument('issue', help='Issue to operate on')
    cmd.add_argument('text', nargs='*', help='Comment text')

    cmd = parser.command('edit', help='Edit issue description or summary', handler=edit_issue)
    cmd.add_argument('issue', help='Issue')
    cmd.add_argument('text', nargs='*', help='New text')

    cmd = parser.command('field', help='Update field values for an issue', handler=issue_fields)
    cmd.add_argument('issue', help='Issue')
    cmd.add_argument('operation', help='Operation', choices=['add', 'set', 'remove', 'sub'])
    cmd.add_argument('name', help='Name of field to update')
    cmd.add_argument('values', help='Value(s) to update', nargs='*')

    cmd = parser.command('fields', help='List fields (and allowed values, when applicable)', handler=issue_fields)
    cmd.add_argument('-t', '--type', default=None, help='Fields available at creation time for the specified type')
    cmd.add_argument('issue', help='Existing Issue (more fields available here)', nargs='?')

    cmd = parser.command('close', help='Move issue(s) to closed/done/resolved', handler=close_issues)
    cmd.add_argument('--subtasks', help='Close subtasks, too', default=False, action='store_true')
    cmd.add_argument('-r', '--resolution', help='Set resolution on transition')
    cmd.add_argument('target', nargs='+', help='Target issue(s)')

    cmd = parser.command('call-api', help='Call an API directly and print the resulting JSON', handler=call_api)
    cmd.add_argument('--raw', help='Produce raw JSON instead of a Python object', default=False, action='store_true')
    cmd.add_argument('resource', help='Location sans host/REST version (e.g. self, issue/KEY-123')

    cmd = parser.command('component', help='Modify component(s)', handler=component_ops)
    cmd.add_argument('-a', '--add', help='Component to add', nargs='+')
    cmd.add_argument('-r', '--remove', help='Component to remove', nargs=1)

    cmd = parser.command('components', help='List components', handler=component_list)
    add_list_options(cmd, quiet_help='Just print component names')
    cmd.add_argument('-s', '--search', help='Search by regular expression')

    cmd = parser.command('template', help='Create issue from YAML template', handler=create_from_template)
    cmd.add_argument('template_file', help='Path to the template file')
    cmd.add_argument('--apply', help='Apply template to existing issue')
    cmd.add_argument('-n', '--non-interactive', default=False, help='Do not prompt for variables', action='store_true')
    cmd.add_argument('-q', '--quiet', default=False, help='Only print new issue IDs after creation (for scripting)', action='store_true')
    cmd.add_argument('--dry-run', default=False, help='Print template with variables substituted; do not file issues', action='store_true')
    cmd.add_argument('vars', help='Variables/values (name value name2 value2 ...)', nargs='*')

    cmd = parser.command('validate', help='Validate a YAML template for use with the "template" command',
                         handler=validate_template)
    cmd.add_argument('template_file', help='Path to the template file')

    cmd = parser.command('generate-template', help='Generate YAML template from existing issue', handler=generate_template)
    cmd.add_argument('issue_id', help='Issue to generate the template from', type=str.upper)
    cmd.add_argument('-a', '--all-fields', default=False, help='Include all fields, even ones that may not make sense for a '
                                                               'template',
                     action='store_true')

    cmd = parser.command('sprint', help='Get Sprint information', handler=sprint_info)
    cmd.add_argument('sprint_id', help='If present, display issues in specific sprint', type=int, nargs='?')
    cmd.add_argument('--all-types', '-A', help='When displaying sprint issues, include subtasks', action='store_true', default=False)
    cmd.add_argument('--new', help='When displaying issues, only list issues not in progress', default=False, action='store_true')
    cmd.add_argument('--raw', '-r', help='When displaying issues, include this additional JQL snippet')
    add_list_options(cmd)
    cmd.add_argument('--closed', help='Include closed sprints or issues', default=False, action='store_true')

    cmd = parser.command('eausm-vote', help='Apply your EZ Agile Planning vote', handler=eausm_vote)
    cmd.add_argument('issue_id', nargs='+', help='Target issue(s)', type=str.upper)
    cmd.add_argument('vote', help='Story Point Value', type=str.upper)

    cmd = parser.command('vote', help='Apply your vote to an issue', handler=vote)
    cmd.add_argument('issue_id', nargs='+', help='Target issue(s)', type=str.upper)
    cmd.add_argument('-r', '--remove', default=False, action='store_true', help='Remove vote from issue(s)')

    return parser


def update_args(args):
    # Special cases
    if len(args) == 3:
        if args[1] == 'field' and args[2] != '-h':
            return [args[0], 'fields', args[2]]
    if len(args) == 4:
        if args[1] == 'field' and '-t' in args:
            return [args[0], 'fields', args[2], args[3]]
    return args


def main():
    parser = create_parser()
    args = update_args(sys.argv)
    ns = parser.parse_args(args=args[1:])

    field = None
    if ns.x_format_field:
        field = ns.x_format_field

    try:
        project = get_jira_project(ns.project, config_file=ns.config, field=field)
    except KeyError:
        print('Configuration faile failed to parse correctly')
        sys.exit(1)
    except FileNotFoundError:
        print('Please create a configuration file (~/.jirate.json)')
        sys.exit(1)
    except Exception as err:
        print(err)
        sys.exit(1)

    # Pass this down in namespace to callbacks
    parser.add_arg('project', project)
    try:
        rc = parser.finalize(ns)
    except JIRAError as err:
        print(err)
        if ns.debug:
            project.request_cache.debug_dump()
        sys.exit(1)
    except Exception as err:
        print(err)
        sys.exit(1)
    if rc:
        ret = rc[0]
        save = rc[1]  # NOQA
    else:
        print('No command specified')
        ret = 0
        save = False  # NOQA

    project.request_cache.save()
    if ns.debug:
        project.request_cache.debug_dump()
    sys.exit(ret)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

from jirate.jboard import Jirate
from jirate.tests import fake_jira
from jirate.jira_fields import apply_field_renderers, render_field_data

import os
import time

import pytest

os.environ['TZ'] = 'America/New_York'
time.tzset()

fake_jirate = Jirate(fake_jira())
test_fielddefs = fake_jirate.jira.fields()


def test_rendering_basic():
    issue = fake_jirate.issue('TEST-1')
    apply_field_renderers()

    assert render_field_data('assignee', issue.raw['fields'], False, False) == ('Assignee', 'Rory Obert - robert@pie.com')


def test_rendering_display():
    # Override assignee to show user name in JIRA
    # Calls predefined function
    fielddefs = [{'id': 'assignee', 'name': 'Assignee', 'display': 'name'}]

    issue = fake_jirate.issue('TEST-1')
    apply_field_renderers(fielddefs)

    assert render_field_data('assignee', issue.raw['fields'], False, False) == ('Assignee', 'rory')


def test_render_nocode_oops():
    # Override assignee to show user name in JIRA
    # Ignores code _and_ display when overriding a base thing, resulting in the dict being produced
    fielddefs = [{'id': 'assignee', 'name': 'Assignee', 'code': 'field[\'emailAddress\']'}]

    issue = fake_jirate.issue('TEST-1')
    apply_field_renderers(fielddefs)

    ret = render_field_data('assignee', issue.raw['fields'], False, False)
    assert ret[0] == 'Assignee'
    assert ret[1] == str(issue.field('assignee'))


def test_render_code():
    # Test code callback
    fielddefs = [{'id': 'assignee', 'name': 'Assignee', 'code': 'field[\'emailAddress\']'}]

    issue = fake_jirate.issue('TEST-1')
    apply_field_renderers(fielddefs)

    assert render_field_data('assignee', issue.raw['fields'], False, True) == ('Assignee', 'robert@pie.com')


def test_render_code_override():
    # Display overrides code when both are present even when saying to allow code
    fielddefs = [{'id': 'assignee', 'name': 'Assignee', 'display': 'name', 'code': 'field[\'emailAddress\']'}]

    issue = fake_jirate.issue('TEST-1')
    apply_field_renderers(fielddefs)

    assert render_field_data('assignee', issue.raw['fields'], False, True) == ('Assignee', 'rory')


field_test_params = 'field_id,field_name,string_value,object_value'
field_test_info = [
    pytest.param('components', 'Component/s', '"food, pork", "food, carrot"', ['food, pork', 'food, carrot']),
    # Fixed in build (string)
    pytest.param('customfield_1234567', 'Fixed in Build', 'test-build-1', 'test-build-1'),

    # Score
    pytest.param('customfield_1234568', 'Score', '22', 22),

    # Array of Options (value) - includes checkboxes
    pytest.param('customfield_1234569', 'Array of Options', 'option_one, option_two', ['option_one', 'option_two']),

    # Array of Versions (names),
    pytest.param('customfield_1234570', 'Array of Versions', 'Version1, Version2', ['Version1', 'Version2']),

    # Array of Users (user names),
    # TODO: Support setting arrays of users by email address
    pytest.param('customfield_1234571', 'Array of Users', 'user-one, two@other.eml', ['user-one', 'two@other.eml']),

    # Array of strings
    pytest.param('customfield_1234572', 'Array of Strings', 'one, two, "three, and four"', ['one', 'two', 'three, and four']),

    # Array of groups (name?),
    pytest.param('customfield_1234573', 'Array of Groups', 'group1, group2', ['group1', 'group2']),

    # Array of Any values
    pytest.param('customfield_1234574', 'Any Value', 'one, 2.0', ['one', 2.0]),

    # Date value
    pytest.param('customfield_1234575', 'Date Value', '2022/08/01', '2022/08/01'),

    # Datetime value
    pytest.param('customfield_1234576', 'Datetime Value', '2019/12/24 9:10 PM', '2019/12/24 9:10 PM'),

    # Related Issue (issue key)
    pytest.param('customfield_1234577', 'Related Issue', 'TEST-2', 'TEST-2'),

    # Option (value),
    pytest.param('customfield_1234578', 'Option Value', 'option_one', 'option_one'),

    # Option with child
    pytest.param('customfield_1234579', 'Option and Child', 'option_one - child_value', 'option_one - child_value'),

    # User value (Name - email address),
    pytest.param('customfield_1234580', 'User Value', 'Rory Obert - robert@pie.com', 'Rory Obert - robert@pie.com'),

    # Version value (name),
    pytest.param('customfield_1234581', 'Version Value', 'Version1', 'Version1'),
]


@pytest.mark.parametrize(field_test_params, field_test_info)
def test_render_auto_defs(field_id, field_name, string_value, object_value):
    apply_field_renderers(test_fielddefs)
    issue = fake_jirate.issue('TEST-1')
    fields = issue.raw['fields']

    # Fixed in build
    assert render_field_data(field_id, fields, False, True) == (field_name, string_value)
    assert render_field_data(field_id, fields, False, as_object=True) == (field_name, object_value)


@pytest.mark.parametrize(field_test_params, field_test_info)
def test_render_null_auto_defs(field_id, field_name, string_value, object_value):
    apply_field_renderers(test_fielddefs)
    issue = fake_jirate.issue('TEST-2')
    fields = issue.raw['fields']

    assert render_field_data(field_id, fields, False, True) == (field_name, None)
    assert render_field_data(field_id, fields, False, True, as_object=True) == (field_name, None)


def test_render_no_field():
    # Not all fields are in the schema
    apply_field_renderers(test_fielddefs)
    issue = fake_jirate.issue('TEST-2')
    fields = issue.raw['fields']

    assert render_field_data('summary', fields, False, True) == ('summary', 'Test 1')
    assert render_field_data('summary', fields, False, True, as_object=True) == ('summary', 'Test 1')


def test_votes_simple():
    apply_field_renderers(test_fielddefs)
    issue = fake_jirate.issue('TEST-1')
    fields = issue.raw['fields']

    assert render_field_data('votes', fields, False, True) == ('Votes', '42')


def test_votes_with_users():
    apply_field_renderers(test_fielddefs)
    issue = fake_jirate.issue('TEST-2')
    fields = issue.raw['fields']

    assert render_field_data('votes', fields, False, True) == ('Votes', '2: Cow, Pig')


def test_watches_simple():
    apply_field_renderers(test_fielddefs)
    issue = fake_jirate.issue('TEST-1')
    fields = issue.raw['fields']

    assert render_field_data('watches', fields, False, True) == ('Watchers', '37')


def test_watches_with_users():
    apply_field_renderers(test_fielddefs)
    issue = fake_jirate.issue('TEST-2')
    fields = issue.raw['fields']

    assert render_field_data('watches', fields, False, True) == ('Watchers', '2: Chicken, Duck')


def test_user_defined_field():
    user_email = [{'name': 'Assignee Email', 'id': 'email', '_jirate_reference': 'assignee', 'code': 'field[\'emailAddress\']'}]

    apply_field_renderers(test_fielddefs)
    apply_field_renderers(user_email)

    issue = fake_jirate.issue('TEST-1')
    fields = issue.raw['fields']

    assert render_field_data('email', fields, False, True) == ('Assignee Email', 'robert@pie.com')



#!/usr/bin/env python

from jirate.tests import fake_metadata, fake_transitions, fake_fields
from jirate.jira_input import transmogrify_input
from jirate.jira_input import check_value, allowed_value_validate

import os
import time

import pytest

os.environ['TZ'] = 'America/New_York'
time.tzset()


def test_trans_string_reflective():
    # identity
    inp = {'description': 'This is a description'}
    out = {'description': 'This is a description'}

    assert transmogrify_input(fake_metadata, **inp) == out


def test_trans_string_nym():
    # We translate to JIRA field IDs
    inp = {'Description': 'This is a description'}
    out = {'description': 'This is a description'}

    assert transmogrify_input(fake_metadata, **inp) == out


def test_trans_priority():
    inp = {'Priority': 'blocker'}
    out = {'priority': {'name': 'Blocker'}}

    assert transmogrify_input(fake_metadata, **inp) == out


def test_trans_custom_basic():
    inp = {'fixed_in_build': 'build123'}
    out = {'customfield_1234567': 'build123'}

    assert transmogrify_input(fake_metadata, **inp) == out


def test_trans_number():
    inp = {'score': '1'}
    out = {'customfield_1234568': 1.0}

    assert transmogrify_input(fake_metadata, **inp) == out


def test_trans_array_options():
    inp = {'array_of_options': 'One,Two'}
    out = {'customfield_1234569': [{'value': 'One'}, {'value': 'Two'}]}

    assert transmogrify_input(fake_metadata, **inp) == out


def test_trans_array_missing_option():
    with pytest.raises(ValueError):
        transmogrify_input(fake_metadata, **{'option_value': 'One,Two,Three'})


def test_trans_array_versions():
    inp = {'array_of_versions': '1.0.1,1.0.2'}
    out = {'customfield_1234570': [{'name': '1.0.1'}, {'name': '1.0.2'}]}

    assert transmogrify_input(fake_metadata, **inp) == out


def test_trans_array_users():
    inp = {'array_of_users': 'user1,user2'}
    out = {'customfield_1234571': [{'name': 'user1'}, {'name': 'user2'}]}

    assert transmogrify_input(fake_metadata, **inp) == out


def test_trans_array_strings():
    inp = {'array_of_strings': '"string one","string 2"'}
    out = {'customfield_1234572': ['string one', 'string 2']}

    assert transmogrify_input(fake_metadata, **inp) == out


def test_trans_array_groups():
    inp = {'array_of_groups': 'group1,group2'}
    out = {'customfield_1234573': [{'name': 'group1'}, {'name': 'group2'}]}

    assert transmogrify_input(fake_metadata, **inp) == out


def test_trans_array_any():
    inp = {'any_value': 'one,2'}
    out = {'customfield_1234574': 'one,2'}

    assert transmogrify_input(fake_metadata, **inp) == out


# TODO: Fixup date and datetime; these should take various input formats
# but are strings


def test_trans_issuelink():
    inp = {'related_issue': 'TEST-2'}
    out = {'customfield_1234577': {'name': 'TEST-2'}}


def test_trans_option():
    inp = {'option_value': 'one'}
    out = {'customfield_1234578': {'value': 'One'}}

    assert transmogrify_input(fake_metadata, **inp) == out


def test_trans_missing_option():
    with pytest.raises(ValueError):
        transmogrify_input(fake_metadata, **{'option_value': 'Four'})


def test_trans_missing_field():
    inp = {'not_a_real_field': 'one'}
    out = {}

    assert transmogrify_input(fake_metadata, **inp) == out


def test_trans_user_value():
    inp = {'User Value': 'user1'}
    out = {'customfield_1234580': {'name': 'user1'}}

    assert transmogrify_input(fake_metadata, **inp) == out


def test_trans_version_value():
    inp = {'version_value': '1.0.1'}
    out = {'customfield_1234581': {'name': '1.0.1'}}

    assert transmogrify_input(fake_metadata, **inp) == out


def test_trans_missing_version():
    with pytest.raises(ValueError):
        transmogrify_input(fake_metadata, **{'version_value': '999'})


def test_trans_metadata():
    transition_fields = fake_transitions


def test_check_value_simple():
    valid = {'1': ['1', ' 1', '1 ', '1-', '-1'],
             'a': ['a', ' a', 'a ', 'a-', '-a'],
             '1.1': [' 1.1', '1.1 '],
             'test': ['test', 'test two']}

    for check in valid:
        for value in valid[check]:
            assert(check_value(check, value))


def test_check_value_exact():
    assert(check_value('1', '1') == 2)


def test_check_value_inexact():
    assert(check_value('1', '1 ') == 1)


def test_check_value_nomatch():
    valid = {'1': ['a1', ' 12', '1.1'],
             'a': ['a1', ' ab', 'a.b'],
             '1.1': ['1.11', '1.1.2']}

    for check in valid:
        for value in valid[check]:
            assert(not check_value(check, value))


def test_fuzzy_values_exact():
    allowed_values = fake_fields[3]['allowedValues']  # components
    # python and 'python 4.0' are available, but because this is exact,
    # we get only one python result
    assert allowed_value_validate('components', 'python', allowed_values) == 'python'


def test_fuzzy_values_fail():
    allowed_values = fake_fields[3]['allowedValues']  # components
    # 2 possible matches for 'test', so ambiguous, raise valueerror
    with pytest.raises(ValueError):
        allowed_value_validate('components', 'test', allowed_values)


def test_fuzzy_values_approx():
    allowed_values = fake_fields[3]['allowedValues']  # components
    # only one match for this
    assert allowed_value_validate('components', 'fuzzy', allowed_values) == 'fuzzy match'


def test_avv_multiple():
    allowed_values = fake_fields[3]['allowedValues']  # components

    assert allowed_value_validate('components', ['fuzzy', 'python'], allowed_values) == ['fuzzy match', 'python']



#!/usr/bin/python3
from jirate.jira_custom import sprint_content_to_py, sprint_field

import pytest  # NOQA


def test_jira_sprint_parse_simple():
    val1 = "com.atlassian.greenhopper.service.sprint.Sprint@49e59daf[id=13133,rapidViewId=21164,state=CLOSED,name=Porky Sprint 5 - Sept,startDate=2024-09-01T06:22:00.000Z,endDate=2024-09-29T06:22:00.000Z,completeDate=2024-10-02T06:38:15.491Z,activatedDate=2024-09-03T02:47:34.778Z,sequence=13133,goal=,synced=false,autoStartStop=false,incompleteIssuesDestinationId=<null>]"

    expected = [{'_hash': '49e59daf',
                 'id': 13133,
                 'rapidViewId': 21164,
                 'state': 'CLOSED',
                 'startDate': '2024-09-01T06:22:00.000Z',
                 'endDate': '2024-09-29T06:22:00.000Z',
                 'completeDate': '2024-10-02T06:38:15.491Z',
                 'name': 'Porky Sprint 5 - Sept',
                 'activatedDate': '2024-09-03T02:47:34.778Z',
                 'sequence': 13133,
                 'goal': '',
                 'synced': False,
                 'autoStartStop': False,
                 'incompleteIssuesDestinationId': None}]

    assert sprint_content_to_py(val1) == expected


def test_sprint_field():
    val1 = "com.atlassian.greenhopper.service.sprint.Sprint@49e59daf[id=13133,rapidViewId=21164,state=CLOSED,name=Porky Sprint 5 - Sept,startDate=2024-09-01T06:22:00.000Z,endDate=2024-09-29T06:22:00.000Z,completeDate=2024-10-02T06:38:15.491Z,activatedDate=2024-09-03T02:47:34.778Z,sequence=13133,goal=,synced=false,autoStartStop=false,incompleteIssuesDestinationId=<null>]"

    assert sprint_field(val1, None, False) == 'Porky Sprint 5 - Sept (ID: 13133)'


def test_sprint_multi_open():
    # Take last active sprint
    val1 = ["com.atlassian.greenhopper.service.sprint.Sprint@56814b6f[id=21301,rapidViewId=21064,state=CLOSED,name=Homies Sprint 5,startDate=2025-01-06T03:00:00.000Z,endDate=2025-01-25T04:00:00.000Z,completeDate=2025-01-25T04:00:43.237Z,activatedDate=2025-01-06T03:00:45.655Z,sequence=21301,goal=<null>,synced=false,autoStartStop=true,incompleteIssuesDestinationId=18234]", "com.atlassian.greenhopper.service.sprint.Sprint@638afb2[id=18234,rapidViewId=21064,state=ACTIVE,name=Homies Sprint 6,startDate=2025-01-27T03:00:00.000Z,endDate=2025-02-15T03:00:00.000Z,completeDate=<null>,activatedDate=2025-01-27T03:00:31.194Z,sequence=18234,goal=<null>,synced=false,autoStartStop=true,incompleteIssuesDestinationId=40133]"]

    assert sprint_field(val1, None, False) == 'Homies Sprint 6 (ID: 18234)'


def test_sprint_multi_closed():
    # Take last closed sprint
    val1 = ["com.atlassian.greenhopper.service.sprint.Sprint@56814b6f[id=21301,rapidViewId=21064,state=CLOSED,name=Homies Sprint 5,startDate=2025-01-06T03:00:00.000Z,endDate=2025-01-25T04:00:00.000Z,completeDate=2025-01-25T04:00:43.237Z,activatedDate=2025-01-06T03:00:45.655Z,sequence=21301,goal=<null>,synced=false,autoStartStop=true,incompleteIssuesDestinationId=18234]", "com.atlassian.greenhopper.service.sprint.Sprint@638afb2[id=18234,rapidViewId=21064,state=CLOSED,name=Homies Sprint 6,startDate=2025-01-27T03:00:00.000Z,endDate=2025-02-15T03:00:00.000Z,completeDate=<null>,activatedDate=2025-01-27T03:00:31.194Z,sequence=18234,goal=<null>,synced=false,autoStartStop=true,incompleteIssuesDestinationId=40133]"]

    assert sprint_field(val1, None, False) == 'Homies Sprint 6 (ID: 18234)'


def test_sprint_multi_future():
    # Take last active sprint
    val1 = ["com.atlassian.greenhopper.service.sprint.Sprint@56814b6f[id=21301,rapidViewId=21064,state=ACTIVE,name=Homies Sprint 5,startDate=2025-01-06T03:00:00.000Z,endDate=2025-01-25T04:00:00.000Z,completeDate=2025-01-25T04:00:43.237Z,activatedDate=2025-01-06T03:00:45.655Z,sequence=21301,goal=<null>,synced=false,autoStartStop=true,incompleteIssuesDestinationId=18234]", "com.atlassian.greenhopper.service.sprint.Sprint@638afb2[id=18234,rapidViewId=21064,state=FUTURE,name=Homies Sprint 6,startDate=2025-01-27T03:00:00.000Z,endDate=2025-02-15T03:00:00.000Z,completeDate=<null>,activatedDate=2025-01-27T03:00:31.194Z,sequence=18234,goal=<null>,synced=false,autoStartStop=true,incompleteIssuesDestinationId=40133]"]

    assert sprint_field(val1, None, False) == 'Homies Sprint 5 (ID: 21301)'


def test_sprint_multi_active():
    # two sprints, I guess?
    val1 = ["com.atlassian.greenhopper.service.sprint.Sprint@56814b6f[id=21301,rapidViewId=21064,state=ACTIVE,name=Homies Sprint 5,startDate=2025-01-06T03:00:00.000Z,endDate=2025-01-25T04:00:00.000Z,completeDate=2025-01-25T04:00:43.237Z,activatedDate=2025-01-06T03:00:45.655Z,sequence=21301,goal=<null>,synced=false,autoStartStop=true,incompleteIssuesDestinationId=18234]", "com.atlassian.greenhopper.service.sprint.Sprint@638afb2[id=18234,rapidViewId=21064,state=ACTIVE,name=Homies Sprint 6,startDate=2025-01-27T03:00:00.000Z,endDate=2025-02-15T03:00:00.000Z,completeDate=<null>,activatedDate=2025-01-27T03:00:31.194Z,sequence=18234,goal=<null>,synced=false,autoStartStop=true,incompleteIssuesDestinationId=40133]"]

    assert sprint_field(val1, None, False) == 'Homies Sprint 5 (ID: 21301), Homies Sprint 6 (ID: 18234)'



#!/usr/bin/python3
#
# A lot of JIRA data is static - or close enough. So, we can
# greatly improve performance by caching a lot of commonly-retrieved
# fields.
#
# Previously, Jirate would simply cache individual items.
# This implementation uses patterns to cache specific API endpoints,
# including pagination, for things that don't change terribly often.
#
# We cache individual API calls with their parameters and the results,
# then return the results if the result is within _expire_time of the
# original request.  By recording the states locally, user-facing performance
# is dramatically improved.
#
# Future improvements:
# - A future improvement might be to allow individual expire times for
#   various patterns that change more frequently.  For example, fetching
#   an issue might have a 1 minute, while fetching /field might be 30 days.
# - Allow a user to purge cache from CLI (besides removing the file)
#
import os
import re
import time
import types

from jira.client import ResilientSession

from .decor import hbar_over, hbar_under
from .localstate import pickle_write, pickle_read

default_cache_expire = 43200

default_cache_patterns = {
    'GET': [r'/rest/api/[0-9]+/myself$',
            r'/rest/api/[0-9]+/field$',
            r'/rest/api/[0-9]+/user/search',
            r'/rest/agile/[0-9]+(\.[0-9]+)?/board$',
            r'/rest/agile/[0-9]+(\.[0-9]+)?/board/[0-9]+/sprint$',
            r'/rest/api/[0-9]+/project/[A-Z]+/statuses$',
            r'/rest/api/[0-9]+/project/[A-Z]+$',
            r'/rest/api/[0-9]+/issue/[0-9]+/transitions',
            r'/rest/api/[0-9]+/issue/createmeta/[A-Z]+/issuetypes',
            r'/rest/api/[0-9]+/issue/[0-9]+/editmeta',
            r'/rest/api/[0-9]+/user\?(username|key)=']
}


def _cached_request(cache, method, url, **kwargs):
    # Check cache first
    ret = cache._cache_read(method, url, kwargs)
    if ret:
        return ret

    # Record cache miss for API profiling/debugging
    cache._dbg_request(method, url, **kwargs)
    ret = ResilientSession.request(cache.session, method, url, **kwargs)
    cache._record_info(method, url, kwargs, ret)
    return ret


class RequestCache(object):
    __req_magic__ = '__req_magic__'

    def __init__(self, session, filename=None, expire=None, **kwargs):
        if expire is None:
            expire = default_cache_expire
        self._expire_time = expire
        self._cache_hits = 0
        self._cache_file = filename
        self.cached_reqs = {'magic': self.__req_magic__, 'GET': {}}
        self.cache_patterns = default_cache_patterns
        self.debug_reqs = {}
        self.user_breaks = {}

        self.session = session
        session.request = types.MethodType(_cached_request, self)
        if not filename:
            return
        self.load(filename)

    def _cache_read(self, method, url, args_dict=None):
        if method not in self.cached_reqs:
            return None
        if url not in self.cached_reqs[method]:
            return None

        for item in self.cached_reqs[method][url]:
            if time.gmtime(item['expire']) <= time.gmtime():
                # Expired
                self.cached_reqs[method][url].remove(item)
                return None
            if args_dict != item['args']:
                continue
            self._cache_hits = self._cache_hits + 1
            return item['value']
        return None

    def _dbg_request(self, method, url, **kwargs):
        if method in self.user_breaks and url in self.user_breaks[method]:
            raise Exception(f'User break @ {method} {url}')
        if method not in self.debug_reqs:
            self.debug_reqs[method] = {}
        if url not in self.debug_reqs[method]:
            self.debug_reqs[method][url] = {'count': 1}
        else:
            self.debug_reqs[method][url]['count'] += 1

    def debug_dump(self):
        total = 0
        print()
        if self.debug_reqs:
            hbar_under('API profiling starts')
            for key in self.debug_reqs:
                print(key)
                for url in self.debug_reqs[key]:
                    count = self.debug_reqs[key][url]['count']
                    total = total + count
                    print(f'    {count} {url}')
        hbar_over(f'Total reqs: {total} Cache hits: {self._cache_hits}')

    def _record_info(self, method, url, args_dict, value):
        if method not in self.cache_patterns:
            return
        urlmatch = False
        for line in self.cache_patterns[method]:
            if re.search(line, url):
                urlmatch = True
                break
        if not urlmatch:
            return
        expire = time.time() + float(self._expire_time)
        if url not in self.cached_reqs[method]:
            self.cached_reqs[method][url] = []
        self.cached_reqs[method][url].append({'args': args_dict,
                                              'expire': expire,
                                              'value': value})

    def load(self, filename=None):
        if filename is None:
            filename = self._cache_file
        if filename is None:
            return None
        # We expire per-req, not per file
        try:
            ret = pickle_read(filename)
        except:  # NOQA
            # Unpickling error, read error, whatever;
            # result is same for our purposes
            ret = False
            pass

        if ret:
            # Verify structure
            if 'magic' not in ret or ret['magic'] != self.__req_magic__:
                ret = None

        if ret:
            self.cached_reqs = ret
            return True
        try:
            os.unlink(filename)
        except FileNotFoundError:
            pass
        return False

    def flush(self):
        for method in self.cached_reqs:
            if method == 'magic':
                continue
            url_nuke = []
            for url in self.cached_reqs[method]:
                nuke = []
                for item in self.cached_reqs[method][url]:
                    if time.gmtime(item['expire']) <= time.gmtime():
                        # Expired
                        nuke.append(item)
                for item in nuke:
                    self.cached_reqs[method][url].remove(item)
                if not self.cached_reqs[method][url]:
                    url_nuke.append(url)
            for url in url_nuke:
                del self.cached_reqs[method][url]

    def save(self, filename=None):
        if filename is None:
            filename = self._cache_file
        if not filename:
            return None
        self.flush()
        return pickle_write(filename, self.cached_reqs)



#!/usr/bin/env python

from jirate.tests import fake_jira, fake_user, fake_transitions, fake_metadata, fake_fields
from jirate.args import GenericArgs
from jirate.jira_cli import _parse_creation_args, _create_from_template, _serialize_issue, _generate_template, \
    validate_template
from jirate.jboard import JiraProject
from jirate.jira_fields import apply_field_renderers

import pytest  # NOQA
import types
from jsonschema.exceptions import ValidationError
from pathlib import Path


fake_jirate = JiraProject(fake_jira(), 'TEST', closed_status='Done', allow_code=True)

# These are used in the template tests below
args = GenericArgs()
args.project = fake_jirate
TEMPLATE_DIR = Path('jirate/tests/templates').absolute()


# XXX This metadata shouldn't just be fields? Fix later.
def _fake_metadata(obj, issuetype, **args):
    return {'fields': fake_metadata}


fake_jirate.issue_metadata = types.MethodType(_fake_metadata, fake_jirate)
apply_field_renderers(fake_fields)


def test_creation_args_bare():
    # With no required/reserved/translated fields, input == output
    issue_data = {'subtasks': {'1': 2}, 'issue_type': 'whatever', 'pork': 'bacon'}

    assert _parse_creation_args(issue_data) == issue_data


def test_creation_args_nominal():
    # Ensure having a translated field and one field that is reserved that
    # we strip out the reserved field and translate the other one.
    reserved_fields = ['subtasks']
    translate_fields = {'issue_type': 'issuetype'}

    issue_data = {'subtasks': {'1': 2}, 'issue_type': 'whatever', 'pork': 'bacon'}
    expected = {'issuetype': 'whatever', 'pork': 'bacon'}

    assert _parse_creation_args(issue_data, reserved_fields=reserved_fields, translate_fields=translate_fields) == expected


def test_creation_args_no_trans():
    # Ensure having a translated field but no instance of it, we don't add extra info
    reserved_fields = ['subtasks']
    translate_fields = {'issue_type': 'issuetype'}

    issue_data = {'subtasks': {'1': 2}, 'pork': 'bacon'}
    expected = {'pork': 'bacon'}

    assert _parse_creation_args(issue_data, reserved_fields=reserved_fields, translate_fields=translate_fields) == expected


def test_creation_args_missing():
    # Given a missing required field in our issue_data, ensure that we
    # raise an error.
    reserved_fields = ['subtasks']
    translate_fields = {'issue_type': 'issuetype'}
    required_fields = ['summary']

    issue_data = {'subtasks': {'1': 2}, 'issue_type': 'whatever', 'pork': 'bacon'}

    with pytest.raises(ValueError):
        assert _parse_creation_args(issue_data, required_fields=required_fields, reserved_fields=reserved_fields, translate_fields=translate_fields)


def test_creation_args_required_trans():
    # Given an untranslated field name in our issue_data, ensure that it is
    # translated.
    reserved_fields = ['subtasks']
    translate_fields = {'issue_type': 'issuetype'}
    required_fields = ['summary', 'issuetype']

    issue_data = {'subtasks': {'1': 2}, 'issue_type': 'whatever', 'pork': 'bacon', 'summary': 'Summary'}
    expected = {'issuetype': 'whatever', 'pork': 'bacon', 'summary': 'Summary'}

    assert _parse_creation_args(issue_data, required_fields=required_fields, reserved_fields=reserved_fields, translate_fields=translate_fields) == expected


def test_creation_args_required_no_trans():
    # Given an input with a required field that is translatable but
    # already in the right format, we don't alter the output.
    reserved_fields = ['subtasks']
    translate_fields = {'issue_type': 'issuetype'}
    required_fields = ['summary', 'issuetype']

    issue_data = {'subtasks': {'1': 2}, 'issuetype': 'whatever', 'pork': 'bacon', 'summary': 'Summary'}
    expected = {'issuetype': 'whatever', 'pork': 'bacon', 'summary': 'Summary'}

    assert _parse_creation_args(issue_data, required_fields=required_fields, reserved_fields=reserved_fields, translate_fields=translate_fields) == expected


def test_creation_args_defaults():
    # Ensure that when we are provided default inputs, that those are copied
    # to our output.  Anything in start_vals should not be removed or ignored
    # if they are in reserved_fields, since start_vals is expected to be
    # code and issue_data is user input.
    reserved_fields = ['subtasks']
    required_fields = ['summary', 'issuetype']
    start_vals = {'issuetype': 'whatever', 'other': 'value'}

    issue_data = {'subtasks': {'1': 2}, 'pork': 'bacon', 'summary': 'Summary'}
    expected = {'pork': 'bacon', 'summary': 'Summary', 'issuetype': 'whatever', 'other': 'value'}

    assert _parse_creation_args(issue_data, required_fields=required_fields, reserved_fields=reserved_fields, start_vals=start_vals) == expected


def test_creation_args_defaults_no_override():
    # Ensure that when we are provided default inputs, that those are copied
    # to our output.  Additionally, when we have a reserved field that is
    # already provided in the start_vals but then again in issue_data, if
    # it is a reserved field, it should not be overridden.
    reserved_fields = ['subtasks', 'issuetype']
    required_fields = ['summary', 'issuetype']
    start_vals = {'issuetype': 'whatever', 'other': 'value'}

    issue_data = {'subtasks': {'1': 2}, 'pork': 'bacon', 'summary': 'Summary', 'issuetype': 'other_issuetype'}
    expected = {'pork': 'bacon', 'summary': 'Summary', 'issuetype': 'whatever', 'other': 'value'}

    assert _parse_creation_args(issue_data, required_fields=required_fields, reserved_fields=reserved_fields, start_vals=start_vals) == expected


def test_creation_args_defaults_override():
    # Ensure that when we are provided default inputs, that those are copied
    # to our output.  Additionally, when we have a field in our start_vals
    # that is not reserved and also provided in issue_data, we should
    # overwrite it.
    reserved_fields = ['subtasks']
    required_fields = ['summary', 'issuetype']
    start_vals = {'issuetype': 'whatever', 'other': 'value'}

    issue_data = {'subtasks': {'1': 2}, 'pork': 'bacon', 'summary': 'Summary', 'issuetype': 'other_issuetype'}
    expected = {'pork': 'bacon', 'summary': 'Summary', 'issuetype': 'other_issuetype', 'other': 'value'}

    assert _parse_creation_args(issue_data, required_fields=required_fields, reserved_fields=reserved_fields, start_vals=start_vals) == expected


def test_create_from_template_simple():
    template = {'issues': [
                {'summary': 'whatever',
                 'issuetype': 'Task',
                 'description': 'Description'}
                ]}

    issue = _create_from_template(args, template)
    assert issue == [{'parent': 'TEST-1'}]
    issue = fake_jirate.issue('TEST-1')
    assert issue == {'key': 'TEST-1', 'raw': {'fields': {'description': 'Description', 'issuetype': 'Task', 'project': {'key': 'TEST'}, 'summary': 'whatever'}}}


def test_create_from_template_customfield():
    # TEST-2 pops out because we are reusing the above fake_jirate
    template = {'issues': [
                {'summary': 'custom field test',
                 'issuetype': 'Task',
                 'fixed_in_build': 'abc123',
                 'description': 'Test of custom field transmogrify'}
                ]}

    issue = _create_from_template(args, template)
    assert issue == [{'parent': 'TEST-2'}]
    issue = fake_jirate.issue('TEST-2')
    assert issue == {'key': 'TEST-2', 'raw': {'fields': {'description': 'Test of custom field transmogrify', 'issuetype': 'Task', 'customfield_1234567': 'abc123', 'project': {'key': 'TEST'}, 'summary': 'custom field test'}}}


def test_create_from_template_subtasks():
    # TEST-3/4 pop out because we are reusing the above fake_jirate
    template = {'issues': [
                {'summary': 'Sub Tasks Check',
                 'issuetype': 'Task',
                 'subtasks': [
                     {'summary': 'Child Task'}
                 ]}]}

    _create_from_template(args, template)
    # Creates TEST-3 and TEST-4
    assert fake_jirate.issue('TEST-3') == {'key': 'TEST-3', 'raw': {'fields': {'issuetype': 'Task', 'project': {'key': 'TEST'}, 'summary': 'Sub Tasks Check'}}}
    assert fake_jirate.issue('TEST-4') == {'key': 'TEST-4', 'raw': {'fields': {'issuetype': 'Sub-task', 'parent': {'key': 'TEST-3'}, 'project': {'key': 'TEST'}, 'summary': 'Child Task'}}}


def test_create_from_template_multiple_types():
    # TEST-5/6 pop out because we are reusing the above fake_jirate
    template = {'issues': [
                {'summary': 'Multitype1',
                 'issuetype': 'Task'
                 },
                {'summary': 'Multitype2',
                 'issuetype': 'Bug',
                 'subtasks': [
                     {'summary': 'Bug Subtask'}
                 ]}]}

    _create_from_template(args, template)
    # Creates TEST-5 and TEST-6
    assert fake_jirate.issue('TEST-5') == {'key': 'TEST-5', 'raw': {'fields': {'issuetype': 'Task', 'project': {'key': 'TEST'}, 'summary': 'Multitype1'}}}
    assert fake_jirate.issue('TEST-6') == {'key': 'TEST-6', 'raw': {'fields': {'issuetype': 'Bug', 'project': {'key': 'TEST'}, 'summary': 'Multitype2'}}}
    assert fake_jirate.issue('TEST-7') == {'key': 'TEST-7', 'raw': {'fields': {'issuetype': 'Sub-task', 'project': {'key': 'TEST'}, 'parent': {'key': 'TEST-6'}, 'summary': 'Bug Subtask'}}}


def test_generate_template_simple():
    # Note this uses the fake_issues found in __init__.py, not from the create_from_template tests
    actual = {'issues': [
                {'summary': 'Test 1',
                 'issue_type': 'Bug',
                 'description': 'Test Description 1'}
                ]}
    allow_fields = list(actual['issues'][0].keys())
    generated = _generate_template(fake_jira().issue('TEST-1').raw['fields'],
                                   fake_jirate.field_to_alias,
                                   allow_fields=allow_fields)
    generated = {'issues': [generated]}

    assert actual == generated


def test_generate_template_customfield():
    # Note this uses the fake_issues found in __init__.py, not from the create_from_template tests
    actual = {'issues': [
                {'summary': 'Test 1',
                 'issue_type': 'Bug',
                 'description': 'Test Description 1',
                 'fixed_in_build': 'test-build-1'}
                ]}
    allow_fields = list(actual['issues'][0].keys())

    generated = _generate_template(fake_jira().issue('TEST-1').raw['fields'],
                                   fake_jirate.field_to_alias,
                                   allow_fields=allow_fields)
    generated = {'issues': [generated]}
    assert actual == generated


def test_generate_template_subtasks():
    # Note this uses the fake_issues found in __init__.py, not from the create_from_template tests
    actual = {'issues': [
                {'summary': 'Test 3 (parent task)',
                 'issue_type': 'Bug',
                 'sub_tasks': [
                     {'summary': 'Test 4 (subtask of Test 3)',
                      'issue_type': 'Sub-task'}
                 ]}]}
    allow_fields = list(actual['issues'][0].keys())

    generated = _generate_template(fake_jira().issue('TEST-3').raw['fields'],
                                   fake_jirate.field_to_alias,
                                   allow_fields=allow_fields,
                                   fetch_issue=fake_jira().issue)
    generated = {'issues': [generated]}

    assert actual == generated


def test_validate_templates_good():
    # Validating a known-good template should succeed
    args.template_file = f"{TEMPLATE_DIR / 'good-template.yaml'}"
    validate_template(args)


def test_validate_templates_bad():
    # Validating a known-bad template should fail
    args.template_file = f"{TEMPLATE_DIR / 'bad-template.yaml'}"
    with pytest.raises(ValidationError):
        validate_template(args)



#!/usr/bin/env python

from jira.client import JIRA
from jira.resources import Issue, dict2resource, Project
from jirate.args import GenericArgs
from jirate.decor import pretty_print

testx = 1

fake_user = {'self': 'https://domain.com/rest/api/2/user?username=porkchop', 'key': 'porkchop', 'name': 'porkchop', 'emailAddress': 'porkchop@domain.com', 'avatarUrls': {'48x48': 'https://domain.com/secure/useravatar?avatarId=1', '24x24': 'https://domain.com/secure/useravatar?size=small&avatarId=1', '16x16': 'https://domain.com/secure/useravatar?size=xsmall&avatarId=1',
                                                                                                                                                                          '32x32': 'https://domain.com/secure/useravatar?size=medium&avatarId=1'}, 'displayName': 'Chop Pork', 'active': True, 'deleted': False, 'timeZone': 'America/New_York', 'locale': 'en_US', 'groups': {'size': 9, 'items': []}, 'applicationRoles': {'size': 1, 'items': []}, 'expand': 'groups,applicationRoles'}


fake_fields = [
    # 0 - parent
    {'clauseNames': ['parent', 'Parent'],
        'custom': False,
        'id': 'parent',
        'name': 'Parent',
        'schema': {'system': 'parent', 'type': 'issuelink'}
     },
    # 1 - priority
    {'clauseNames': ['priority', 'Priority'],
        'custom': False,
        'id': 'priority',
        'name': 'Priority',
        'schema': {'system': 'priority', 'type': 'priority'},
        'allowedValues': [
            {'id': 101,
             'name': 'Blocker'},
            {'id': 102,
             'name': 'Critical'},
            {'id': 103,
             'name': 'Major'},
            {'id': 104,
             'name': 'Normal'},
            {'id': 105,
             'name': 'Minor'},
            {'id': 106,
             'name': 'Undefined'}],
        'defaultValue': {'id': 104, 'name': 'Normal'},
        'hasDefaultValue': True,
        'operations': ['set']
     },
    # 2 - description
    {'clauseNames': ['description', 'Description'],
        'custom': False,
        'name': 'Description',
        'id': 'description',
        'required': True,
        'operations': ['set'],
        'schema': {'type': 'string'},
     },
    # 3 - components
    {'clauseNames': ['components', 'Component/s'],
        'custom': False,
        'id': 'components',
        'name': 'Component/s',
        'schema': {'system': 'items', 'type': 'array', 'items': 'component'},
        'allowedValues': [
            {'id': 1001,
             'name': 'python'},
            {'id': 1002,
             'name': 'kernel'},
            {'id': 1003,
             'name': 'glibc'},
            {'id': 1004,
             'name': 'porkchop'},
            {'id': 1005,
             'name': 'test two'},
            {'id': 1006,
             'name': 'test three'},
            {'id': 1007,
             'name': 'fuzzy match'},
            {'id': 1008,
             'name': 'python 4.0'}],
        'operations': ['add', 'set', 'remove'],
        'required': True
     },
    {'clauseNames': ['cf[1234567]', 'Fixed in Build'],
        'custom': True,
        'id': 'customfield_1234567',
        'name': 'Fixed in Build',
        'navigable': True,
        'orderable': True,
        'schema': {'custom': 'com.atlassian.jira.plugin.system.customfieldtypes:textfield',
                   'customId': 1234567,
                   'type': 'string'},
        'operations': ['set'],
        'searchable': True},
    {'clauseNames': ['cf[1234568]', 'Score'],
        'custom': True,
        'id': 'customfield_1234568',
        'name': 'Score',
        'navigable': True,
        'orderable': True,
        'schema': {'custom': 'org.jboss.labs.jira.plugin.jboss-custom-field-types-plugin:jbonlynumber',
                   'customId': 1234568,
                   'type': 'number'},
        'operations': ['set'],
        'searchable': True},
    {'clauseNames': ['cf[1234569]', 'Array of Options'],
        'custom': True,
        'id': 'customfield_1234569',
        'name': 'Array of Options',
        'navigable': True,
        'orderable': True,
        'schema': {'custom': 'com.atlassian.jira.plugin.system.customfieldtypes:multiselect',
                   'customId': 1234569,
                   'items': 'option',
                   'type': 'array'},
        'allowedValues': [
            {'value': 'One',
             'id': '1'},
            {'value': 'Two',
             'id': '2'},
            {'value': 'Three',
             'id': '3'}
    ],
        'operations': ['add', 'set', 'remove'],
        'searchable': True},
    {'clauseNames': ['cf[1234570]', 'Array of Versions'],
        'custom': True,
        'id': 'customfield_1234570',
        'name': 'Array of Versions',
        'navigable': True,
        'orderable': True,
        'schema': {'custom': 'com.atlassian.jira.plugin.system.customfieldtypes:multiselect',
                   'customId': 1234570,
                   'items': 'version',
                   'type': 'array'},
        'allowedValues': [
            {'name': '1.0.1',
             'id': '11'},
            {'name': '1.0.2',
             'id': '12'}
    ],
        'operations': ['add', 'set', 'remove'],
        'searchable': True},
    {'clauseNames': ['cf[1234571]', 'Array of Users'],
        'custom': True,
        'id': 'customfield_1234571',
        'name': 'Array of Users',
        'navigable': True,
        'orderable': True,
        'schema': {'custom': 'com.atlassian.jira.plugin.system.customfieldtypes:multiselect',
                   'customId': 1234571,
                   'items': 'user',
                   'type': 'array'},
        'operations': ['add', 'set', 'remove'],
        'searchable': True},
    {'clauseNames': ['cf[1234572]', 'Array of Strings'],
        'custom': True,
        'id': 'customfield_1234572',
        'name': 'Array of Strings',
        'navigable': True,
        'orderable': True,
        'schema': {'custom': 'com.atlassian.jira.plugin.system.customfieldtypes:multiselect',
                   'customId': 1234572,
                   'items': 'string',
                   'type': 'array'},
        'operations': ['add', 'set', 'remove'],
        'searchable': True},
    {'clauseNames': ['cf[1234573]', 'Array of Groups'],
        'custom': True,
        'id': 'customfield_1234573',
        'name': 'Array of Groups',
        'navigable': True,
        'orderable': True,
        'schema': {'custom': 'com.atlassian.jira.plugin.system.customfieldtypes:multiselect',
                   'customId': 1234573,
                   'items': 'group',
                   'type': 'array'},
        'searchable': True},
    {'clauseNames': ['cf[1234574]', 'Any Value'],
        'custom': True,
        'id': 'customfield_1234574',
        'name': 'Any Value',
        'navigable': True,
        'orderable': True,
        'schema': {'custom': 'com.atlassian.jira.plugin.system.customfieldtypes:multiselect',
                   'customId': 1234574,
                   'type': 'any'},
        'operations': ['add', 'set', 'remove'],
        'searchable': True},
    {'clauseNames': ['cf[1234575]', 'Date Value'],
        'custom': True,
        'id': 'customfield_1234575',
        'name': 'Date Value',
        'navigable': True,
        'orderable': True,
        'schema': {'custom': 'com.atlassian.jira.plugin.system.customfieldtypes:multiselect',
                   'customId': 1234575,
                   'type': 'date'},
        'operations': ['set'],
        'searchable': True},
    {'clauseNames': ['cf[1234576]', 'Datetime Value'],
        'custom': True,
        'id': 'customfield_1234576',
        'name': 'Datetime Value',
        'navigable': True,
        'orderable': True,
        'schema': {'custom': 'com.atlassian.jira.plugin.system.customfieldtypes:multiselect',
                   'customId': 1234576,
                   'type': 'datetime'},
        'operations': ['set'],
        'searchable': True},
    {'clauseNames': ['cf[1234577]', 'Related Issue'],
        'custom': True,
        'id': 'customfield_1234577',
        'name': 'Related Issue',
        'navigable': True,
        'orderable': True,
        'schema': {'custom': 'com.atlassian.jira.plugin.system.customfieldtypes:multiselect',
                   'customId': 1234577,
                   'type': 'issuelinks'},
        'operations': ['set'],
        'searchable': True},
    {'clauseNames': ['cf[1234578]', 'Option Value'],
        'custom': True,
        'id': 'customfield_1234578',
        'name': 'Option Value',
        'navigable': True,
        'orderable': True,
        'operations': ['set'],
        'allowedValues': [
            {'value': 'One',
             'id': '1'},
            {'value': 'Two',
             'id': '2'},
            {'value': 'Three',
             'id': '3'}
    ],
        'schema': {'custom': 'com.atlassian.jira.plugin.system.customfieldtypes:multiselect',
                   'customId': 1234578,
                   'type': 'option'},
        'searchable': True},
    {'clauseNames': ['cf[1234579]', 'Option and Child'],
        'custom': True,
        'id': 'customfield_1234579',
        'name': 'Option and Child',
        'navigable': True,
        'orderable': True,
        'schema': {'custom': 'com.atlassian.jira.plugin.system.customfieldtypes:multiselect',
                   'customId': 1234579,
                   'type': 'option-with-child'},
        'allowedValues': [
            {'value': 'One',
             'id': '1'},
            {'value': 'Two',
             'id': '2'},
            {'value': 'Three',
             'id': '3'}
    ],
        'searchable': True},
    {'clauseNames': ['cf[1234580]', 'User Value'],
        'custom': True,
        'id': 'customfield_1234580',
        'name': 'User Value',
        'navigable': True,
        'orderable': True,
        'schema': {'custom': 'com.atlassian.jira.plugin.system.customfieldtypes:multiselect',
                   'customId': 1234580,
                   'type': 'user'},
        'searchable': True},
    {'clauseNames': ['cf[1234581]', 'Version Value'],
        'custom': True,
        'id': 'customfield_1234581',
        'name': 'Version Value',
        'navigable': True,
        'orderable': True,
        'schema': {'custom': 'com.atlassian.jira.plugin.system.customfieldtypes:multiselect',
                   'customId': 1234581,
                   'type': 'version'},
        'allowedValues': [
            {'name': '1.0.1',
             'id': '11'},
            {'name': '1.0.2',
             'id': '12'}
    ],
        'searchable': True},
      {
        'id': 'summary',
        'name': 'Summary',
        'custom': False,
        'orderable': True,
        'navigable': True,
        'searchable': True,
        'clauseNames': [
          'summary'
        ],
        'schema': {
          'type': 'string',
          'system': 'summary'
        }
      },
      {
        'id': 'issuetype',
        'name': 'Issue Type',
        'custom': False,
        'orderable': True,
        'navigable': True,
        'searchable': True,
        'clauseNames': [
          'issuetype',
          'type'
        ],
        'schema': {
          'type': 'issuetype',
          'system': 'issuetype'
        }
      },
      {
        'id': 'subtasks',
        'name': 'Sub-Tasks',
        'custom': False,
        'orderable': False,
        'navigable': True,
        'searchable': False,
        'clauseNames': [
          'subtasks'
        ],
        'schema': {
          'type': 'array',
          'items': 'issuelinks',
          'system': 'subtasks'
        }
  }

]


fake_metadata = {val['id']: val for val in fake_fields}


fake_transitions = {'expand': 'transitions',
                    'transitions': [{'id': '11', 'name': 'New', 'to': {'name': 'New', 'id': '10000'}},
                                    {'id': '12', 'name': 'In Progress', 'to': {'name': 'In Progres', 'id': '10001'}},
                                    {'id': '13', 'name': 'Done', 'to': {'name': 'Done', 'id': '10002'},
                                     'fields':
                                         {'resolution': {
                                             'name': 'Resolution',
                                             'fieldId': 'resolution',
                                             'schema': {
                                                 'type': 'resolution',
                                                 'system': 'resolution'},
                                             'operations': ['set'],
                                             'allowedValues': [{'name': 'Done'}, {'name': 'Won\'t Do'}, {'name': 'Duplicate'}]}}},
                                    {'id': '14', 'name': 'New', 'to': {'name': 'New', 'id': '10003'}}]
                    }


fake_issues = {
    'TEST-1': {'expand': 'renderedFields,names,schema,operations,editmeta,changelog,versionedRepresentations',
               'fields': {'aggregateprogress': {'progress': 0, 'total': 0},
                          'aggregatetimeestimate': None,
                          'aggregatetimeoriginalestimate': None,
                          'aggregatetimespent': None,
                          'archivedby': None,
                          'archiveddate': None,
                          'assignee': {'active': True,
                                       'displayName': 'Rory Obert',
                                       'emailAddress': 'robert@pie.com',
                                       'key': 'rory',
                                       'name': 'rory',
                                       'self': 'https://domain.com/rest/api/2/user?username=rory',
                                       'timeZone': 'America/New_York'},
                          'attachment': [],
                          'comment': {'comments': []},
                          'components': [{'name': 'food, pork'}, {'name': 'food, carrot'}],
                          'created': '2023-08-03T10:28:48.366+0000',
                          'description': 'Test Description 1',
                          'duedate': '2024-01-30',
                          'environment': None,
                          'fixVersions': [],
                          'issuelinks': [],
                          'issuetype': {'avatarId': 13263,
                                        'description': 'A bug',
                                        'id': '1',
                                        'name': 'Bug',
                                        'self': 'https://domain.com/rest/api/2/issuetype/1',
                                        'subtask': False},
                          'labels': ['label1', 'label2'],
                          'lastViewed': '2023-11-22T15:51:50.281+0000',
                          'priority': {'id': '104', 'name': 'Normal'},
                          'progress': {'progress': 0, 'total': 0},
                          'project': {'id': '1234',
                                      'key': 'TEST',
                                      'name': 'Test Project',
                                      'projectCategory': {'description': 'Test Projects',
                                                          'id': '1073',
                                                          'name': 'Projecto',
                                                          'self': 'https://domain.com/rest/api/2/projectCategory/1073'},
                                      'projectTypeKey': 'software',
                                      'self': 'https://domain.com/rest/api/2/project/1234'},
                          'reporter': {'active': True,
                                       'displayName': 'Pork Chop',
                                       'emailAddress': 'pchop@domai.com',
                                       'key': 'JIRAUSER1',
                                       'name': 'pchop@domain.com',
                                       'self': 'https://domain.com/rest/api/2/user?username=pchop%40domain.com',
                                       'timeZone': 'Asia/Kolkata'},
                          'resolution': None,
                          'resolutiondate': None,
                          'security': None,
                          'status': {'description': 'Initial creation status. ',
                                     'id': '10',
                                     'name': 'New',
                                     'self': 'https://domain.com/rest/api/2/status/10016',
                                     'statusCategory': {'colorName': 'default',
                                                        'id': 2,
                                                        'key': 'new',
                                                        'name': 'To Do',
                                                        'self': 'https://domain.com/rest/api/2/statuscategory/2'}},
                          'subtasks': [],
                          'summary': 'Test 1',
                          'timeestimate': None,
                          'timeoriginalestimate': None,
                          'timespent': None,
                          'timetracking': {},
                          'updated': '2023-11-30T15:06:39.875+0000',
                          'versions': [{'archived': False,
                                        'description': '',
                                        'id': '6',
                                        'name': 'version-6',
                                        'released': False,
                                        'self': 'https://domain.com/rest/api/2/version/6'}],
                          'votes': {'hasVoted': False,
                                    'self': 'https://domain.com/rest/api/2/issue/TEST-1/votes',
                                    'votes': 42},
                          'watches': {'isWatching': False,
                                      'self': 'https://domain.com/rest/api/2/issue/TEST-1/watchers',
                                      'watchCount': 37},
                          'worklog': {'maxResults': 20,
                                      'startAt': 0,
                                      'total': 0,
                                      'worklogs': []},
                          'workratio': -1,
                          'customfield_1234567': 'test-build-1',
                          'customfield_1234568': 22.0,
                          'customfield_1234569': [{'name': 'Option1', 'value': 'option_one'},
                                                  {'name': 'Option2', 'value': 'option_two'}],
                          'customfield_1234570': [{'name': 'Version1', 'value': 'version_one'},
                                                  {'name': 'Version2', 'value': 'version_two'}],
                          'customfield_1234571': [{'key': 'user1', 'name': 'user-one', 'displayName': 'One', 'emailAddress': 'one@two.com'},
                                                  {'key': 'user2', 'name': 'two@other.eml', 'displayName': 'Two', 'emailAddress': 'two@two.com'}],
                          'customfield_1234572': ['one', 'two', 'three, and four'],
                          'customfield_1234573': [{'name': 'group1'}, {'name': 'group2'}],
                          'customfield_1234574': ['one', 2.0],
                          'customfield_1234575': '2022-08-01',
                          'customfield_1234576': '2019-12-25T02:10:00.000+0000',
                          'customfield_1234577': {'key': 'TEST-2'},
                          'customfield_1234578': {'name': 'Option1', 'value': 'option_one'},
                          'customfield_1234579': {'name': 'Option1', 'value': 'option_one', 'child': {'name': 'child1', 'value': 'child_value'}},
                          'customfield_1234580': {
                   'displayName': 'Rory Obert',
                                  'emailAddress': 'robert@pie.com',
                                  'key': 'rory',
                                  'name': 'rory',
                                  'self': 'https://domain.com/rest/api/2/user?username=rory',
                                  'timeZone': 'America/New_York'},
                          'customfield_1234581': {'name': 'Version1', 'value': 'version_one'}
                          },
               'id': '1000001',
               'key': 'TEST-1',
               'self': 'https://domain.com/rest/api/2/issue/1000001'},

    'TEST-2': {'expand': 'renderedFields,names,schema,operations,editmeta,changelog,versionedRepresentations',
               'fields': {'aggregateprogress': {'progress': 0, 'total': 0},
                          'aggregatetimeestimate': None,
                          'aggregatetimeoriginalestimate': None,
                          'aggregatetimespent': None,
                          'archivedby': None,
                          'archiveddate': None,
                          'assignee': None,
                          'attachment': [],
                          'comment': {'comments': []},
                          'components': [],
                          'created': '2023-08-03T10:28:48.366+0000',
                          'customfield_1234567': None,
                          'customfield_1234568': None,
                          'customfield_1234569': None,
                          'description': 'Test Description 2',
                          'duedate': '2024-01-30',
                          'environment': None,
                          'fixVersions': [],
                          'issuelinks': [],
                          'issuetype': {'avatarId': 13263,
                                        'description': 'A bug',
                                        'id': '1',
                                        'name': 'Bug',
                                        'self': 'https://domain.com/rest/api/2/issuetype/1',
                                        'subtask': False},
                          'labels': ['label2'],
                          'lastViewed': '2023-11-22T15:51:50.281+0000',
                          'priority': {'id': 104, 'name': 'Normal'},
                          'progress': {'progress': 0, 'total': 0},
                          'project': {'id': '1234',
                                      'key': 'TEST',
                                      'name': 'Test Project',
                                      'projectCategory': {'description': 'Test Projects',
                                                          'id': '1073',
                                                          'name': 'Projecto',
                                                          'self': 'https://domain.com/rest/api/2/projectCategory/1073'},
                                      'projectTypeKey': 'software',
                                      'self': 'https://domain.com/rest/api/2/project/1234'},
                          'reporter': {'active': True,
                                       'displayName': 'Pork Chop',
                                       'emailAddress': 'pchop@domai.com',
                                       'key': 'JIRAUSER1',
                                       'name': 'pchop@domain.com',
                                       'self': 'https://domain.com/rest/api/2/user?username=pchop%40domain.com',
                                       'timeZone': 'Asia/Kolkata'},
                          'resolution': None,
                          'resolutiondate': None,
                          'security': None,
                          'status': {'description': 'Initial creation status. ',
                                     'id': '10',
                                     'name': 'New',
                                     'self': 'https://domain.com/rest/api/2/status/10016',
                                     'statusCategory': {'colorName': 'default',
                                                        'id': 2,
                                                        'key': 'new',
                                                        'name': 'To Do',
                                                        'self': 'https://domain.com/rest/api/2/statuscategory/2'}},
                          'subtasks': [],
                          'summary': 'Test 1',
                          'timeestimate': None,
                          'timeoriginalestimate': None,
                          'timespent': None,
                          'timetracking': {},
                          'updated': '2023-11-30T15:06:39.875+0000',
                          'versions': [{'archived': False,
                                        'description': '',
                                        'id': '6',
                                        'name': 'version-6',
                                        'released': False,
                                        'self': 'https://domain.com/rest/api/2/version/6'}],
                          'votes': {'hasVoted': False,
                                    'self': 'https://domain.com/rest/api/2/issue/TEST-1/votes',
                                    'votes': 2, 'voters': [{'displayName': 'Cow'}, {'displayName': 'Pig'}]},
                          'watches': {'isWatching': False,
                                      'self': 'https://domain.com/rest/api/2/issue/TEST-1/watchers',
                                      'watchCount': 2, 'watchers': [{'displayName': 'Chicken'}, {'displayName': 'Duck'}]},
                          'worklog': {'maxResults': 20,
                                      'startAt': 0,
                                      'total': 0,
                                      'worklogs': []},
                          'workratio': -1},
               'id': '1000002',
               'key': 'TEST-2',
               'self': 'https://domain.com/rest/api/2/issue/1000002'},
    'TEST-3': {'expand': 'renderedFields,names,schema,operations,editmeta,changelog,versionedRepresentations',
               'fields': {'aggregateprogress': {'progress': 0, 'total': 0},
                          'aggregatetimeestimate': None,
                          'aggregatetimeoriginalestimate': None,
                          'aggregatetimespent': None,
                          'archivedby': None,
                          'archiveddate': None,
                          'assignee': None,
                          'attachment': [],
                          'comment': {'comments': []},
                          'components': [],
                          'created': '2023-08-03T10:28:48.366+0000',
                          'customfield_1234567': None,
                          'customfield_1234568': None,
                          'customfield_1234569': None,
                          'description': 'Test Description 3',
                          'duedate': '2024-01-30',
                          'environment': None,
                          'fixVersions': [],
                          'issuelinks': [],
                          'issuetype': {'avatarId': 13263,
                                        'description': 'A bug',
                                        'id': '1',
                                        'name': 'Bug',
                                        'self': 'https://domain.com/rest/api/2/issuetype/1',
                                        'subtask': False},
                          'parent': {},
                          'labels': ['label2'],
                          'lastViewed': '2023-11-22T15:51:50.281+0000',
                          'priority': {'id': 104, 'name': 'Normal'},
                          'progress': {'progress': 0, 'total': 0},
                          'project': {'id': '1234',
                                      'key': 'TEST',
                                      'name': 'Test Project',
                                      'projectCategory': {'description': 'Test Projects',
                                                          'id': '1073',
                                                          'name': 'Projecto',
                                                          'self': 'https://domain.com/rest/api/2/projectCategory/1073'},
                                      'projectTypeKey': 'software',
                                      'self': 'https://domain.com/rest/api/2/project/1234'},
                          'reporter': {'active': True,
                                       'displayName': 'Pork Chop',
                                       'emailAddress': 'pchop@domai.com',
                                       'key': 'JIRAUSER1',
                                       'name': 'pchop@domain.com',
                                       'self': 'https://domain.com/rest/api/2/user?username=pchop%40domain.com',
                                       'timeZone': 'Asia/Kolkata'},
                          'resolution': None,
                          'resolutiondate': None,
                          'security': None,
                          'status': {'description': 'Initial creation status. ',
                                     'id': '10',
                                     'name': 'New',
                                     'self': 'https://domain.com/rest/api/2/status/10016',
                                     'statusCategory': {'colorName': 'default',
                                                        'id': 2,
                                                        'key': 'new',
                                                        'name': 'To Do',
                                                        'self': 'https://domain.com/rest/api/2/statuscategory/2'}},
                          'subtasks':
                              [
                                  {
                                    'id': '1000004',
                                    'key': 'TEST-4',
                                    'self': 'https://domain.com/rest/api/2/issue/1000004',
                                    'fields': {
                                      'summary': 'Test 4 (subtask of Test 3)',
                                      'status': {'description': 'Initial creation status. ',
                                                 'id': '10',
                                                 'name': 'New',
                                                 'self': 'https://domain.com/rest/api/2/status/10016',
                                                 'statusCategory': {'colorName': 'default',
                                                                    'id': 2,
                                                                    'key': 'new',
                                                                    'name': 'To Do',
                                                                    'self': 'https://domain.com/rest/api/2/statuscategory/2'}
                                                 },
                                      'priority': {'id': 104, 'name': 'Normal'},
                                      'issuetype': {'avatarId': 13276,
                                                    'description': 'The sub-task of the issue',
                                                    'id': '5',
                                                    'name': 'Sub-task',
                                                    'self': 'https://domain.com/rest/api/2/issuetype/5',
                                                    'subtask': True}
                                    }
                                  }
                          ],
                          'summary': 'Test 3 (parent task)',
                          'timeestimate': None,
                          'timeoriginalestimate': None,
                          'timespent': None,
                          'timetracking': {},
                          'updated': '2023-11-30T15:06:39.875+0000',
                          'versions': [{'archived': False,
                                        'description': '',
                                        'id': '6',
                                        'name': 'version-6',
                                        'released': False,
                                        'self': 'https://domain.com/rest/api/2/version/6'}],
                          'votes': {'hasVoted': False,
                                    'self': 'https://domain.com/rest/api/2/issue/TEST-3/votes',
                                    'votes': 0},
                          'watches': {'isWatching': False,
                                      'self': 'https://domain.com/rest/api/2/issue/TEST-3/watchers',
                                      'watchCount': 2},
                          'worklog': {'maxResults': 20,
                                      'startAt': 0,
                                      'total': 0,
                                      'worklogs': []},
                          'workratio': -1},
               'id': '1000003',
               'key': 'TEST-3',
               'self': 'https://domain.com/rest/api/2/issue/1000003'},
    'TEST-4': {'expand': 'renderedFields,names,schema,operations,editmeta,changelog,versionedRepresentations',
               'fields': {'aggregateprogress': {'progress': 0, 'total': 0},
                          'aggregatetimeestimate': None,
                          'aggregatetimeoriginalestimate': None,
                          'aggregatetimespent': None,
                          'archivedby': None,
                          'archiveddate': None,
                          'assignee': None,
                          'attachment': [],
                          'comment': {'comments': []},
                          'components': [],
                          'created': '2023-08-03T10:28:48.366+0000',
                          'customfield_1234567': None,
                          'customfield_1234568': None,
                          'customfield_1234569': None,
                          'description': None,
                          'duedate': '2024-01-30',
                          'environment': None,
                          'fixVersions': [],
                          'issuelinks': [],
                          'issuetype': {'avatarId': 13276,
                                        'description': 'The sub-task of the issue',
                                        'id': '5',
                                        'name': 'Sub-task',
                                        'self': 'https://domain.com/rest/api/2/issuetype/5',
                                        'subtask': True},
                          'parent':  {'id': '1000003',
                                      'key': 'TEST-3',
                                      'self': 'https://domain.com/rest/api/2/issue/1000003',
                                      'fields': {
                                        'summary': 'Test 3 (parent task)',
                                        'status': {'description': 'Initial creation status. ',
                                                   'id': '10',
                                                   'name': 'New',
                                                   'self': 'https://domain.com/rest/api/2/status/10016',
                                                   'statusCategory': {'colorName': 'default',
                                                                      'id': 2,
                                                                      'key': 'new',
                                                                      'name': 'To Do',
                                                                      'self':
                                                                          'https://domain.com/rest/api/2/statuscategory/2'}},
                                        'priority': {'id': 104, 'name': 'Normal'},
                                        'issuetype': {'avatarId': 13263,
                                                      'description': 'A bug',
                                                      'id': '1',
                                                      'name': 'Bug',
                                                      'self': 'https://domain.com/rest/api/2/issuetype/1',
                                                      'subtask': False},
                                      }},
                          'labels': ['label2'],
                          'lastViewed': '2023-11-22T15:51:50.281+0000',
                          'priority': {'id': 104, 'name': 'Normal'},
                          'progress': {'progress': 0, 'total': 0},
                          'project': {'id': '1234',
                                      'key': 'TEST',
                                      'name': 'Test Project',
                                      'projectCategory': {'description': 'Test Projects',
                                                          'id': '1073',
                                                          'name': 'Projecto',
                                                          'self': 'https://domain.com/rest/api/2/projectCategory/1073'},
                                      'projectTypeKey': 'software',
                                      'self': 'https://domain.com/rest/api/2/project/1234'},
                          'reporter': {'active': True,
                                       'displayName': 'Pork Chop',
                                       'emailAddress': 'pchop@domai.com',
                                       'key': 'JIRAUSER1',
                                       'name': 'pchop@domain.com',
                                       'self': 'https://domain.com/rest/api/2/user?username=pchop%40domain.com',
                                       'timeZone': 'Asia/Kolkata'},
                          'resolution': None,
                          'resolutiondate': None,
                          'security': None,
                          'status': {'description': 'Initial creation status. ',
                                     'id': '10',
                                     'name': 'New',
                                     'self': 'https://domain.com/rest/api/2/status/10016',
                                     'statusCategory': {'colorName': 'default',
                                                        'id': 2,
                                                        'key': 'new',
                                                        'name': 'To Do',
                                                        'self': 'https://domain.com/rest/api/2/statuscategory/2'}},
                          'subtasks': [],
                          'summary': 'Test 4 (subtask of Test 3)',
                          'timeestimate': None,
                          'timeoriginalestimate': None,
                          'timespent': None,
                          'timetracking': {},
                          'updated': '2023-11-30T15:06:39.875+0000',
                          'versions': [{'archived': False,
                                        'description': '',
                                        'id': '6',
                                        'name': 'version-6',
                                        'released': False,
                                        'self': 'https://domain.com/rest/api/2/version/6'}],
                          'votes': {'hasVoted': False,
                                    'self': 'https://domain.com/rest/api/2/issue/TEST-4/votes',
                                    'votes': 0},
                          'watches': {'isWatching': False,
                                      'self': 'https://domain.com/rest/api/2/issue/TEST-4/watchers',
                                      'watchCount': 2},
                          'worklog': {'maxResults': 20,
                                      'startAt': 0,
                                      'total': 0,
                                      'worklogs': []},
                          'workratio': -1},
               'id': '1000004',
               'key': 'TEST-4',
               'self': 'https://domain.com/rest/api/2/issue/1000004'}
}


fake_statuses = [
        {'id': '12363',
         'name': 'Bug',
         'subtask': False,
         'statuses': [{'id': '10000', 'name': 'New', 'statusCategory': { 'id': 2, 'name': 'To Do'}},
                      {'id': '10001', 'name': 'In Progress', 'statusCategory': { 'id': 4, 'name': 'In Progress' }},
                      {'id': '10002', 'name': 'Done', 'statusCategory': { 'id': 6, 'name': 'Done'}}
                     ]
         }
        ]

fake_project = {   'archived': False,
    'assigneeType': 'UNASSIGNED',
    'components': [   {   'description': 'Comp1',
                          'id': '12392576',
                          'isAssigneeTypeValid': False,
                          'name': 'aardvark',
                          'self': 'https://issues.pie.com/rest/api/2/component/12392576'},
                      {   'description': 'anteater',
                          'id': '12392393',
                          'isAssigneeTypeValid': False,
                          'name': 'anteater',
                          'self': 'https://issues.pie.com/rest/api/2/component/12392393'},
                      {   'description': 'yellow belly',
                          'id': '12392577',
                          'isAssigneeTypeValid': False,
                          'name': 'marmot',
                          'self': 'https://issues.pie.com/rest/api/2/component/12392577'}],
    'description': 'The Test Project',
    'expand': 'description,lead,url,projectKeys',
    'id': '12336920',
    'issueTypes': [   {
                          'description': '',
                          'iconUrl': 'https://issues.pie.com/secure/viewavatar?size=xsmall&avatarId=13275&avatarType=issuetype',
                          'id': '20',
                          'name': 'Bug',
                          'self': 'https://issues.pie.com/rest/api/2/issuetype/20',
                          'subtask': False},
                      {   'avatarId': 13267,
                          'description': 'Created by Jira Software - do not '
                                         'edit or delete. Issue type for a big '
                                         'user story that needs to be broken '
                                         'down.',
                          'iconUrl': 'https://issues.pie.com/secure/viewavatar?size=xsmall&avatarId=13267&avatarType=issuetype',
                          'id': '16',
                          'name': 'Epic',
                          'self': 'https://issues.pie.com/rest/api/2/issuetype/16',
                          'subtask': False},
                      {   'avatarId': 13275,
                          'description': 'Created by Jira Software - do not '
                                         'edit or delete. Issue type for a '
                                         'user story.',
                          'iconUrl': 'https://issues.pie.com/secure/viewavatar?size=xsmall&avatarId=13275&avatarType=issuetype',
                          'id': '17',
                          'name': 'Story',
                          'self': 'https://issues.pie.com/rest/api/2/issuetype/17',
                          'subtask': False},
                      {   'avatarId': 13278,
                          'description': 'A task that needs to be done.',
                          'iconUrl': 'https://issues.pie.com/secure/viewavatar?size=xsmall&avatarId=13278&avatarType=issuetype',
                          'id': '3',
                          'name': 'Task',
                          'self': 'https://issues.pie.com/rest/api/2/issuetype/3',
                          'subtask': False},
                      {   'avatarId': 13276,
                          'description': 'The sub-task of the issue',
                          'iconUrl': 'https://issues.pie.com/secure/viewavatar?size=xsmall&avatarId=13276&avatarType=issuetype',
                          'id': '5',
                          'name': 'Sub-task',
                          'self': 'https://issues.pie.com/rest/api/2/issuetype/5',
                          'subtask': True}],
    'key': 'TEST',
    'lead': {   'active': True,
                'displayName': "Pork Chop",
                'key': 'pchop',
                'name': 'pchop@pie.com',
                'self': 'https://issues.pie.com/rest/api/2/user?username=pchop@pie.com'},
    'name': 'Test Project',
    'projectCategory': {   'description': 'The Project Category',
                           'id': '10730',
                           'name': 'Category',
                           'self': 'https://issues.pie.com/rest/api/2/projectCategory/10730'},
    'projectTypeKey': 'software',
    'versions': [   {   'archived': False,
                        'description': '1.0',
                        'id': '12416509',
                        'name': 'version-1.0',
                        'overdue': False,
                        'projectId': 12336920,
                        'releaseDate': '2032-01-01',
                        'released': False,
                        'self': 'https://issues.pie.com/rest/api/2/version/12416509',
                        'startDate': '2024-01-01',
                        'userReleaseDate': '2024/06/01',
                        'userStartDate': '2024/01/01'},
                    {   'archived': False,
                        'description': '2.0',
                        'id': '12416345',
                        'name': 'version-2.0',
                        'overdue': False,
                        'projectId': 12336920,
                        'releaseDate': '2024-03-05',
                        'released': False,
                        'self': 'https://issues.pie.com/rest/api/2/version/12416345',
                        'startDate': '2023-12-01',
                        'userReleaseDate': '2024/12/01',
                        'userStartDate': '2024/12/01'}]
}


class fake_jira_session():
    def __init__(self):
        self.get_urls = []
        self.post_urls = {}
        self.delete_urls = []

    def get(self, url):
        self.get_urls.append(url)
        if url.endswith('statuses'):
            return fake_statuses

    def post(self, url, data=None):
        self.post_urls[url] = data

    def delete(self, url):
        self.delete_urls.append(url)

    def close(self):
        pass

    def reset(self):
        self.__init__()


class fake_jira(JIRA):
    def __init__(self, **kwargs):
        self._fields_cache_value = {}
        self.deploymentType = 'Server'
        self._version = (9, 0, 0)
        self._options = {'async': False}

    def _get_url(self, url_fragment, **args):
        pass

    def _get_json(self, url_fragment, **args):
        pass

    def add_simple_link(self, issue, link):
        pass

    def search_users(self, username):
        pass

    def fields(self):
        return fake_fields

    def search_issues(self, seach_query, startAt=None, maxResults=None):
        pass

    def create_issue(self, **args):
        global testx

        ret = GenericArgs()
        proj = args['project']
        ret.key = f'{proj}-{testx}'
        testx = testx + 1
        ret.raw = {'fields': args}
        ret.raw['fields']['project'] = {'key': proj}
        pretty_print(ret)
        return ret

    def issue(self, issue_key):
        if issue_key.upper() not in fake_issues:
            return None
        ret = Issue(None, None)
        ret.raw = fake_issues[issue_key]
        dict2resource(fake_issues[issue_key], ret)
        return ret

    def issue_link_types(self):
        pass

    def create_issue_link(self, text, left_key, right_key):
        pass

    def remote_links(self, issue_key):
        pass

    def remote_link(self, left_key, right_key):
        pass

    def delete_issue_link(self, left_key, right_key):
        pass

    def project(self, project_key):
        ret = Project({'server': 'issues.pie.com', 'rest_path': 'rest/api/2', 'rest_api_version': 2, 'async': False}, None)
        ret.raw = fake_project
        dict2resource(fake_project, ret)
        return ret

    def myself(self):
        return fake_user

    _session = fake_jira_session()



#!/usr/bin/python

import re

from jirate.decor import parse_params
from jirate.decor import nym


def in_string(value):
    return str(value)


def in_number(value):
    return float(value)


def in_key(value):
    return {'key': value}


def in_name(value):
    return {'name': value}


def in_value(value):
    return {'value': value}


def in_owc(value):
    if ':' not in value:
        raise ValueError('Option-with-child is of form "option:child"')
    vals = value.split(':')
    if len(vals) != 2:
        raise ValueError('Option-with-child is of form "option:child"')

    return {'value': vals[0], 'child': {'value': vals[1]}}


# these can functions can be referenced in custom user
# config files.  These should be generic and should not
# utilize custom fields.
_input_renderers = {
    'number': in_number,
    'option': in_value,
    'priority': in_name,
    'version': in_name,
    'securitylevel': in_name,
    'option-with-child': in_owc,
    'issuelink': in_key,
    'resolution': in_name,
    'user': in_name
}


def _input_list(vals, attrib):
    return [{attrib: val} for val in vals]


def in_user_list(vals):
    # Could be key, which is why it's separate
    return _input_list(vals, 'name')


def in_value_list(vals):
    return _input_list(vals, 'value')


def in_name_list(vals):
    return _input_list(vals, 'name')


def in_string_list(vals):
    return vals


_input_array_renderers = {
    'user': in_user_list,
    'option': in_value_list,
    'version': in_name_list,
    'group': in_name_list,
    'component': in_name_list
}


# Custom field inputs for a few things
def in_sprint_field(value):
    if isinstance(value, list):
        value = value[0]
    return in_number(value)


def in_issue_key(value):
    if isinstance(value, list):
        value = value[0]
    return str(value).upper()


_custom_field_input = {
    'com.pyxis.greenhopper.jira:gh-sprint': in_sprint_field,
    'com.atlassian.jpo:jpo-custom-field-parent': in_issue_key,
    'com.pyxis.greenhopper.jira:gh-epic-link': in_issue_key
}


# Because allowed Values can be very long and complicated, we
# want users to be able to provide, essentially, the minimum
# unique value to save keystrokes (and frustration).
def check_value(check, value):
    for possible_value in (value, nym(value), value.lower()):
        if possible_value == check:
            return 2
        if check not in possible_value:
            continue
        # begnning of string, end of string.
        # Values should be divided by a space or dash
        # This is just a guess; we may want this configurable
        #
        rx = f'(^|[\\s\\-]){check}($|[\\s\\-])'
        if re.search(rx, possible_value):
            return 1
    return 0


def allowed_value_validate(field_name, values, allowed_values=None):
    if not allowed_values:
        return values

    # Validate that the name or value exists and create our array of IDs
    # corresponding to them.
    if not isinstance(values, list):
        check = [values]
    else:
        check = values

    info = {}
    for val in check:
        for av in allowed_values:
            # Ignore archived or disabled allowed values
            if 'archived' in av and av['archived']:
                continue
            if 'disabled' in av and av['disabled']:
                continue

            # Not sure why we care about name vs. value; value only is fine
            for key in ['name', 'value']:
                if key not in av:
                    continue
                cv = check_value(val, av[key])
                if cv == 2:
                    info[val] = {'values': [av[key]], 'exact': True}
                    continue
                if cv == 1:
                    if val in info:
                        if not info[val]['exact']:
                            info[val]['values'].append(av[key])
                    else:
                        info[val] = {'values': [av[key]], 'exact': False}

        # If we didn't find a match, raise an error
        if val not in info:
            raise ValueError(f'Value {val} not allowed for {field_name}')

    # We should only have one match per value
    for key in info:
        if len(info[key]['values']) > 1:
            raise ValueError(f'{key}: More than one match for {key}')

    ret = []
    # Output order MUST match input order
    for val in check:
        ret.append(info[val]['values'][0])

    if not isinstance(values, list):
        return ret[0]

    return ret


def transmogrify_value(value, field_info):
    schema = field_info['schema']
    av = field_info['allowedValues'] if 'allowedValues' in field_info else None

    if 'custom' in schema and schema['custom'] in _custom_field_input:
        return _custom_field_input[schema['custom']](value)

    if schema['type'] == 'array':
        vals = allowed_value_validate(field_info['name'], parse_params(value), av)
        try:
            ret = _input_array_renderers[schema['items']](vals)
        except KeyError:
            ret = in_string_list(vals)
    else:
        try:
            out_val = allowed_value_validate(field_info['name'], value, av)
            ret = _input_renderers[schema['type']](out_val)
        except KeyError:
            ret = in_string(value)

    return ret


# Channelling ... Calvin
def transmogrify_input(field_definitions, **args):
    """ Translate text input into something Jira understands natively when
    pasting to the API.

    Parameters:
        field_definitions: Create/Update metadata or /field dictionary
        args: User-provided key,value pairs (strings)

    Returns:
        Updated dictionary with JIRA field IDs and values corresponding to
        metadata in field_definitions
    """
    drop_fields = ['attachment', 'reporter', 'issuelinks']  # These are not set during create/update
    simple_fields = ['project', 'issuetype', 'summary', 'description']                # Don't process these fields at all
    output = {}
    def_map = {}

    for field in field_definitions:
        def_map[field] = field
        def_map[nym(field)] = field
        def_map[field_definitions[field]['name']] = field
        def_map[nym(field_definitions[field]['name'])] = field

    for field in args:
        value = args[field]
        if field in simple_fields:
            output[field] = value
            continue
        if field not in def_map:
            continue
        field_id = def_map[field]
        if field_id in drop_fields:
            continue
        output[field_id] = transmogrify_value(value, field_definitions[field_id])
    return output
